/* $Id: fi_defs.h,v 1.1.1.1 2002/08/16 17:29:27 root Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */
/* Header file for fault injection relevant data structures */

#ifndef FI_DEFS_H
#define FI_DEFS_H
#include "external_constants.h"
#define FI_DBG 0 /* set this to zero to cut down on debugging output */
#define FI_LOG 0 /* set this to zero to cut down on logging output */
/*******************************************************************************
 * constants
 ******************************************************************************/

#define FI_DBG_FILE "fi_dbg"
#define FI_LOG_FILE "fi_log"
#define FI_ERR_FILE "fi_err"

#define BLOCKMEDIA_NAME "media"
#define FI_MAX_FLTD 1024 /* size of array containing fault description structs */
#define FI_MAX_STRLEN 50
#define FI_SBUF_SIZE 1000 /* size reserved for string allocation */

typedef long int typeofword; /* should be 4 byte */
#define SIZEOFWORD_BYTES sizeof(typeofword)
#define SIZEOFWORD_BITS (SIZEOFWORD_BYTES*8)

#define FIEOF       -1
#define FI_EPARSE -2
#define FI_EMULT  -3
#define FI_EOFFSET  -4
#define FI_ENOSUCHDEVICE -5
#define FI_EILLATIME -6
#define FI_EILLFLTD  -7

/* bitmasks */
#define FI_WATCH_OPEN      0x00000001
#define FI_WATCH_LSEEK     0x00000002
#define FI_WATCH_READ      0x00000004
#define FI_WATCH_WRITE     0x00000008
#define FI_WATCH_CLOSE     0x00000010
#define FI_WATCH_SENDTO    0x00000020
#define FI_WATCH_RECVFROM  0x00000040
#define FI_WATCH_BIND      0x00000080
#define FI_WATCH_CONNECT   0x00000100
#define FI_WATCH_SOCKETCALL \
	(FI_WATCH_BIND|FI_WATCH_RECVFROM|FI_WATCH_SENDTO|FI_WATCH_CONNECT)

/* different positive integer values, used as indices into the same array */
#define NONTRANSIENT_FAULT_ANUM           0
#define FAULT_RAM            1
#define FAULT_CPU_REGS       2
#define FAULT_MMU            3
#define FAULT_HD_BLOCKS      4
#define FAULT_HD_FAIL        5
#define FAULT_NET_RECV       6
#define FAULT_NET_SEND       7
#define FAULT_POWER          8
#define FI_MAX_FLT_LOC      10 /* increase for more fault locations */
/********************************/                            
#define FAULT_PERMANENT     21
#define FAULT_TRANSIENT     22
#define FAULT_DURATION      23
#define FAULT_DURATION_END  24
/********************************/                            
#define END_OF_FAULTLOAD    25

#define END_OF_FAULTLOAD_STR "END_OF_FAULTLOAD"

#define FAULT_RAM_STR          "FAULT_RAM"
#define FAULT_CPU_REGS_STR     "FAULT_CPU_REGS"
#define FAULT_MMU_STR          "FAULT_MMU"
                               
#define FAULT_HD_BLOCKS_STR    "FAULT_HD_BLOCKS"
#define FAULT_HD_FAIL_STR      "FAULT_HD_FAIL"
#define FAULT_NET_SEND_STR     "FAULT_NET_SEND"
#define FAULT_NET_RECV_STR     "FAULT_NET_RECV"
                               
#define FAULT_POWER_STR        "FAULT_POWER"

#define FAULT_PERMANENT_STR    "FAULT_PERMANENT"
#define FAULT_TRANSIENT_STR    "FAULT_TRANSIENT"
#define FAULT_DURATION_STR     "FAULT_DURATION"
#define FAULT_DURATION_END_STR "FAULT_DURATION_END"

#define CPU_REG_EAX  1
#define CPU_REG_EBX  2
#define CPU_REG_ECX  3
#define CPU_REG_EDX  4
#define CPU_REG_ESI  5
#define CPU_REG_EDI  6
#define CPU_REG_EBP  7
#define CPU_REG_ESP  8
#define CPU_REG_EIP  9
#define CPU_REG_DS  10
#define CPU_REG_CS  11
#define CPU_REG_SS  12
#define CPU_REG_ES  13
#define CPU_REG_FS  14
#define CPU_REG_GS  15

#define CPU_REG_EAX_STR "eax"
#define CPU_REG_EBX_STR "ebx"
#define CPU_REG_ECX_STR "ecx"
#define CPU_REG_EDX_STR "edx"
#define CPU_REG_ESI_STR "esi"
#define CPU_REG_EDI_STR "edi"
#define CPU_REG_EBP_STR "ebp"
#define CPU_REG_ESP_STR "esp"
#define CPU_REG_EIP_STR "eip"
#define CPU_REG_DS_STR  "ds"
#define CPU_REG_CS_STR  "cs"
#define CPU_REG_SS_STR  "ss"
#define CPU_REG_ES_STR  "es"
#define CPU_REG_FS_STR  "fs"
#define CPU_REG_GS_STR  "gs"

/*******************************************************************************
 * macros
 ******************************************************************************/
/* when adding new peripheral faults don't forget to fix this macro */
#define PERIPHERAL_FAULTS_ACTIVE \
	(0 < fi_.active[FAULT_HD_FAIL] \
	 	+ fi_.active[FAULT_HD_BLOCKS] \
	 	+ fi_.active[FAULT_NET_SEND] \
		+ fi_.active[FAULT_NET_RECV])
#define FI_SC_ANNULLED_ON_ENTRY (-1 < fi_.syscall)
#define FI_SC_NORMAL (-1 == fi_.syscall)
#define FI_SC_SAVE_ORIG_FROM(orig_eax) {\
	assert(-1 == fi_.syscall) ; \
	fi_.syscall = orig_eax ; \
	}
#define FI_SC_RESTORE_ORIG_TO(orig_eax) {\
	orig_eax = fi_.syscall ; \
	fi_.syscall = -1 ; \
	}
/*******************************************************************************
 * structure definitions
 ******************************************************************************/
/*******************************************************************************
 * structure for string allocation
 ******************************************************************************/
/*
typedef struct fi_sbuf {
	char *free ;
	char buf[FI_SBUF_SIZE] ;
} fi_sbuf_t ;
*/
/*******************************************************************************
 * structures for different fault locations
 ******************************************************************************/
struct flt_hd {
	char *device ;
	int fd ; /* file descriptor used to access file implementing disk */
	loff_t block ;
	loff_t range ;
} ;

struct flt_ram {
	unsigned long addr;
	int bit;
};

struct flt_cpu {
	int reg ;
	int bit ;
} ;

struct flt_net {
	char *device ; /* name of device, e.g. eth0 */
	int fd ; /* file descriptor used to access socket implementing device */
	unsigned short port ; /* host port number of socket */
} ;

/*******************************************************************************
 * structure for general fault description
 ******************************************************************************/
typedef struct fltd_s fi_fltd_t ;
struct fltd_s {
	fi_fltd_t *prev ; /* NULL if first fault in list */
	fi_fltd_t *next ; /* NULL if last fault in list */

	int id ; /* for unambiguous identification of fault */
	int location ; /* hd, net, cpu, ram, power */
	char type ; /* permanent, transient, duration */
	/* Permanent faults are activated once and will stay active until 
	 * end of experiment.
	 * Transient faults are injected once and only once.
	 * Duration faults are activated once and will disappear again after the
	 * given duration.
	 */
	unsigned long atime ; /* activation time */
	unsigned long aduration  ; /* duration of activation */
	union {
	 struct flt_cpu  cpu;
	 struct flt_ram  ram;
	 struct flt_hd   hd;
	 struct flt_net  net;
	 /*
	 struct flt_mmu  mmu;
	 struct flt_power power;
	 */
	} fltd ;
} ;

/** FIXME: not needed any more
typedef struct fi_ptarray_s {
	void *a[FI_MAX_FLTD] ;
} fi_ptarray_t ;
**/

typedef struct fd_struct_s {
	loff_t offset; /* so it works for llseek etc. */
	void *buf;
	int cnt;
	char open; /* flag open yes/no */
	void * a[FI_MAX_FLTD];
	char fault; /* flag "this syscall injected fault" yes/no */
} fd_struct_t;

/*******************************************************************************
*/
typedef struct fi_flt_inj {
	int pid ; /* pid of UMKernel */
	unsigned long watch_enter ;  /* bitmask of system calls to watch */
	unsigned long watch_return ;  /* bitmask of system calls to watch */
	int num_fltd ; /* number of fault descriptions */
	fi_fltd_t *watch_open_ptr ;
	int netinit ;
	int fd ;
	long syscall ;
	int socketcall ;
	struct flt_net console ;
	int active[FI_MAX_FLT_LOC]; /* array of number of active faults */
	struct flt_net netinfo[MAX_ETH_CARDS] ;
	fi_fltd_t fdl[FI_MAX_FLTD] ; /* array of fault descriptions */
	fd_struct_t fds[FI_MAX_FLTD] ; /* index is file descriptor */
	int ncpu;
} fi_flt_inj_t ;
/*******************************************************************************
 * global variables
 ******************************************************************************/
#endif
