/* $Id: fi_inject.c,v 1.1.1.1 2002/08/16 17:29:27 root Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <linux/net.h>
#include <netinet/in.h>
#include <sys/syscall.h>
#include <sys/ptrace.h>
#include <sys/user.h>

#include "lilo.h"
#include "config.h"
#include "fi_defs.h"

/***********************************************************************
 * file global variables
 **********************************************************************/
fi_fltd_t *fi_afdl_first;
fi_fltd_t *fi_afdl_last;
fi_flt_inj_t fi_;

/***********************************************************************
 * function prototypes
 ***********************************************************************
 * file internal functions 
 **********************************************************************/
void fltd_aappend(fi_fltd_t **, fi_fltd_t *);
void fltd_aremove(fi_fltd_t **, int);

unsigned long flip_bit(int,unsigned long);
int fi_peek_data(int, void *, void *, int);
int fi_in_range(loff_t, unsigned long,  fi_fltd_t *);

void fi_annull_syscall_enter(struct user_regs_struct *);
void fi_annull_syscall_return(struct user_regs_struct *);

/*
void fi_watch_open_enter(struct user_regs_struct*);
void fi_watch_open_return(struct user_regs_struct*);
*/
void fi_watch_lseek_enter(struct user_regs_struct*);
void fi_watch_lseek_return(struct user_regs_struct*);
void fi_watch_llseek_enter(struct user_regs_struct*);
void fi_watch_llseek_return(struct user_regs_struct*);
void fi_watch_read_enter(struct user_regs_struct*);
void fi_watch_read_return(struct user_regs_struct*);
void fi_watch_write_enter(struct user_regs_struct*);
void fi_watch_write_return(struct user_regs_struct*);
/*
void fi_watch_close_enter(struct user_regs_struct*);
void fi_watch_close_return(struct user_regs_struct*);
*/

/*
void fi_watch_bind_enter(struct user_regs_struct* );
void fi_watch_connect_enter(struct user_regs_struct* );
*/
void fi_watch_socketcall_enter(struct user_regs_struct*);
void fi_watch_socketcall_return(struct user_regs_struct*);
void fi_watch_sendto_enter(struct user_regs_struct*);
void fi_watch_sendto_return(struct user_regs_struct*);
void fi_watch_recvfrom_enter(struct user_regs_struct*);
void fi_watch_recvfrom_return(struct user_regs_struct*);

void inject_ram(fi_fltd_t *) ;
void inject_cpu_regs(fi_fltd_t *);
void fi_inject_trans_fault(fi_fltd_t *);

void inject_power() ;

void send_frontend(char*, ...);
void send_reg_frontend(struct user_regs_struct*, int);
/***********************************************************************
 * exported functions
 **********************************************************************/
void fi_watch_enter(int, struct user_regs_struct*); /* tracer.c */
void fi_watch_return(int, struct user_regs_struct*); /* tracer.c */
void activate_fltd(fi_fltd_t *) ; /* fi.c */

/***********************************************************************
 * imported functions
 **********************************************************************/
extern void fi_dprintf(char *, ...); /* fi_print.c */
extern void fi_lprintf(char *, ...); /* fi_print.c */
extern void fi_dump_regs (struct user_regs_struct *); /* fi_print.c */

void
send_frontend(char* fmt, ...)
{
	va_list ap;
extern int fi_socket ; /* fi.c */
	int socket = fi_socket;
	ssize_t ret;
	size_t len;
#define SENDBUFSIZE 256
	char buf[SENDBUFSIZE] ;
	memset(buf,0,SENDBUFSIZE);

	assert(0<socket);
	va_start(ap, fmt);
	vsnprintf(buf, SENDBUFSIZE-1, fmt, ap);
	len = strlen(buf) ;
	ret = write(socket,buf,len);
	/*
	if (ret != len) {
		fprintf(stderr, "lilo: %s: %d, %d, %d: %s.\n", __FUNCTION__,
				ret, len, errno, strerr(errno));
	}
	*/
	/*
	assert(ret == len);
	*/
	va_end(ap);
}
void
send_reg_frontend(
	struct user_regs_struct *regs,
	int which)
{
	unsigned long reg = 0;
	char *str;
	switch (which) {
	case CPU_REG_EAX:
		reg = (unsigned long) (regs->eax);
		str = CPU_REG_EAX_STR;
		break;
	case CPU_REG_EBX:
		reg = (unsigned long) (regs->ebx);
		str = CPU_REG_EBX_STR;
		break;
	case CPU_REG_ECX:
		reg = (unsigned long) (regs->ecx);
		str = CPU_REG_ECX_STR;
		break;
	case CPU_REG_EDX:
		reg = (unsigned long) (regs->edx);
		str = CPU_REG_EDX_STR;
		break;
	case CPU_REG_ESI:
		reg = (unsigned long) (regs->esi);
		str = CPU_REG_ESI_STR;
		break;
	case CPU_REG_EDI:
		reg = (unsigned long) (regs->edi);
		str = CPU_REG_EDI_STR;
		break;
	case CPU_REG_EBP:
		reg = (unsigned long) (regs->ebp);
		str = CPU_REG_EBP_STR;
		break;
	case CPU_REG_ESP:
		reg = (unsigned long) (regs->esp);
		str = CPU_REG_ESP_STR;
		break;
	case CPU_REG_EIP:
		reg = (unsigned long) (regs->eip);
		str = CPU_REG_EIP_STR;
		break;
	case CPU_REG_DS:
		reg = (unsigned long) (regs->xds);
		str = CPU_REG_DS_STR;
		break;
	case CPU_REG_CS:
		reg = (unsigned long) (regs->xcs);
		str = CPU_REG_CS_STR;
		break;
	case CPU_REG_SS:
		reg = (unsigned long) (regs->xss);
		str = CPU_REG_SS_STR;
		break;
	case CPU_REG_ES:
		reg = (unsigned long) (regs->xes);
		str = CPU_REG_ES_STR;
		break;
	case CPU_REG_FS:
		reg = (unsigned long) (regs->xfs); 
		str = CPU_REG_FS_STR;
		break;
	case CPU_REG_GS:
		reg = (unsigned long) (regs->xgs); 
		str = CPU_REG_GS_STR;
		break;
	default:
		assert(0) ;
		break;
	}
	send_frontend("%s = 0x%08lx\n", str, reg);
}
/***********************************************************************
 * fltd_aappend
 ***********************************************************************
 * Input:     fi_fltd_t **fdl_start  =  pointer to pointer to start of fault
 *                                      description list
 *            fi_fltd_t *fltd_new  =  pointer to new fault description to append
 * Output:    none
 * Effects:   appends fltd_new to active fault descr. list
 *            installs watchpoints for fault injection if necessary
 * Called By: activate_fltd
***********************************************************************/
void
fltd_aappend(
	fi_fltd_t **fdl_start,
	fi_fltd_t *fltd_new)
{
	/*
	fi_dprintf("%s: call(fdl_start=%p, fltd_new=%p)\n", __FUNCTION__,
			fdl_start, fltd_new);
	*/
	int i;

	assert(fdl_start != NULL);
	assert(fltd_new != NULL);

	fi_.active[NONTRANSIENT_FAULT_ANUM]++;
	if(*fdl_start == NULL) {
		*fdl_start = fltd_new;
	}else{
		assert(fi_afdl_last != NULL);
		fi_afdl_last->next = fltd_new;
	}
	fi_afdl_last = fltd_new;
	switch (fltd_new->location){
	case FAULT_HD_BLOCKS:
		fi_.active[FAULT_HD_BLOCKS]++;
		/*
		fi_.watch_enter  |= FI_WATCH_OPEN;
		*/
		fi_.watch_enter  |= FI_WATCH_LSEEK;
		i=0;
		while (fi_.fds[fltd_new->fltd.hd.fd].a[i]!=NULL && i<FI_MAX_FLTD){
			i++;
		}
		assert(i<FI_MAX_FLTD);
		(fi_.fds[fltd_new->fltd.hd.fd]).a[i] = fltd_new;
		break;
	case FAULT_HD_FAIL:
		fi_.active[FAULT_HD_FAIL]++;
		/*
		fi_.watch_enter  |= FI_WATCH_OPEN;
		*/
		fi_.watch_enter  |= FI_WATCH_LSEEK;
		i=0;
		while (fi_.fds[fltd_new->fltd.hd.fd].a[i]!=NULL && i<FI_MAX_FLTD){
			i++;
		}
		assert(i<FI_MAX_FLTD);
		(fi_.fds[fltd_new->fltd.hd.fd]).a[i] = fltd_new;
		break;
	case FAULT_NET_SEND:
		fi_.active[FAULT_NET_SEND]++;
		fi_.watch_enter  |= FI_WATCH_SENDTO;
		/* Put sock-descr. corresponding to network device into
		 * set of descriptors to watch.
		 */
		if(0<=fltd_new->fltd.net.fd) {
			fi_.fds[fltd_new->fltd.net.fd].a[0] = fltd_new;
		}
		break;
	case FAULT_NET_RECV:
		fi_.active[FAULT_NET_RECV]++;
		fi_.watch_enter  |= FI_WATCH_RECVFROM;
		if(0<=fltd_new->fltd.net.fd) {
			fi_.fds[fltd_new->fltd.net.fd].a[1] = fltd_new;
		}
		break;
	default:
		send_frontend("illegal combination %s + %s\n",
				fltd_new->location, fltd_new->type);
		break;
	}
}
/***********************************************************************
 * fltd_aremove
 ***********************************************************************
 * Input:     fi_fltd_t **fdl_start  =  pointer to pointer to start of fault
 *                                      description list
 *            int id  =  id of fault description to remove
 * Output:    none
 * Effects:   removes fault description identified by id from active fault
 *            description list
 *            removes watchpoints for fault injection if necessary
 * Called By: activate_fltd
 **********************************************************************/
void fltd_aremove(
	fi_fltd_t **fdl_start,
	int id)
{
	int i;
	fi_fltd_t *fdl_prev;
	fi_fltd_t *fdl_cur = *fdl_start;

	assert( fdl_start != (void *)0);
	assert(*fdl_start != (fi_fltd_t *)0);

	/* remove start element of list */
	if ((*fdl_start)->id == id) {
		*fdl_start = (*fdl_start)->next;
	} else { /* remove from middle or at end of list */
		fdl_prev = *fdl_start;
		fdl_cur = fdl_prev->next;
		while((fdl_cur != NULL) && (fdl_cur->id != id)) {
			fdl_prev = fdl_cur;
			fdl_cur = fdl_cur->next;
		}
		assert(fdl_cur != NULL) ;
		fdl_prev->next = fdl_cur->next;
	}

	switch(fdl_cur->location){
	case FAULT_HD_BLOCKS:
	case FAULT_HD_FAIL:
		i=0;
		while (fi_.fds[fdl_cur->fltd.hd.fd].a[i]!=NULL
			&& i<FI_MAX_FLTD
			&& fi_.fds[fdl_cur->fltd.hd.fd].a[i]!=fdl_cur){
			i++;
		}
		assert(i<FI_MAX_FLTD);
		assert(fi_.fds[fdl_cur->fltd.hd.fd].a[i]==fdl_cur);
		(fi_.fds[fdl_cur->fltd.hd.fd]).a[i] = NULL;
		while(i+1 < FI_MAX_FLTD
			&& (fi_.fds[fdl_cur->fltd.hd.fd]).a[i+1]!=NULL){
			(fi_.fds[fdl_cur->fltd.hd.fd]).a[i] =
				(fi_.fds[fdl_cur->fltd.hd.fd]).a[i+1];
			i++;
		}
		break;
	case FAULT_NET_SEND:
		if(0<=fdl_cur->fltd.net.fd) {
			fi_.fds[fdl_cur->fltd.net.fd].a[0] = NULL;
		}
		break;
	case FAULT_NET_RECV:
		if(0<=fdl_cur->fltd.net.fd) {
			fi_.fds[fdl_cur->fltd.net.fd].a[1] = NULL;
		}
		break;
	default:
		fi_lprintf("ERROR: unknown fault location %d\n",
				fdl_cur->location);
		assert(0);
		break;
	}

	/* check if any more of faults in this location are active and remove fault
	 * injection watchpoints if necessary.
	 */
	assert (0<=fdl_cur->location && fdl_cur->location<FI_MAX_FLT_LOC);
	fi_.active[NONTRANSIENT_FAULT_ANUM]--;
	fi_.active[fdl_cur->location]--;
	if (fi_.active[fdl_cur->location] == 0 ){
		switch(fdl_cur->location){
		case FAULT_HD_BLOCKS:
		case FAULT_HD_FAIL:
			if (fi_.active[FAULT_HD_BLOCKS]+fi_.active[FAULT_HD_FAIL] == 0) {
				fi_.watch_enter &= ~FI_WATCH_LSEEK;
			}
			break;
		case FAULT_NET_SEND:
			if (fi_.active[FAULT_NET_SEND] == 0) {
				fi_.watch_enter &= ~FI_WATCH_SENDTO;
			}
			break;
		case FAULT_NET_RECV:
			if (fi_.active[FAULT_NET_RECV] == 0) {
				fi_.watch_enter &= ~FI_WATCH_RECVFROM;
			}
			break;
		default:
			send_frontend("ERROR: unknown fault location %d\n",
					fdl_cur->location);
			assert(0);
			break;
		}
	}
}
/***********************************************************************
 * flip_bit
 * Input:     int bit  =  number of bit (counting from zero = rightmost bit)
 *            unsigned long reg  =  value of register
 * Output:    contents of register with bit flipped.
 * Called By: inject_ram, inject_cpu_regs
***********************************************************************/
unsigned long flip_bit(int bit, unsigned long reg)
{
	return (reg ^= 1<<(bit%(SIZEOFWORD_BITS)));
}
/***********************************************************************
 * fi_peek_data
 ***********************************************************************
 * Input:     int pid  =  process id of traced process
 *            void *src  =  memory address of start of data to get
 *            void *dst  =  pointer to buffer large enough to hold received data
 *            int size  =  how many bytes of data to get (must be multiple of
 *                         word size)
 * Effect:    fills *dst with data
 * Return:    -1 on error
 *             0 otherwise
 * Called By: watch_*_enter-functions
 **********************************************************************/
int
fi_peek_data(
	int pid,
	void *src,
	void *dst,
	int size)
{
	int numwords, i;
	typeofword word;

	assert(src != NULL);
	assert(dst != NULL);

	numwords = size/SIZEOFWORD_BYTES;
	assert(size%SIZEOFWORD_BYTES == 0);

	memset(dst, 0, size);
	for(i=0;i<numwords;i++){
		errno = 0;
		word = ptrace(PTRACE_PEEKDATA, pid,
				(void*)((unsigned
				long)src+i*SIZEOFWORD_BYTES), 0);
		if (errno) return -1;

		*((typeofword*)((unsigned long)dst + i*SIZEOFWORD_BYTES)) = word;
#if 0
		fi_dprintf("      src=%08p, dst=%08p, word=%08lx\n",
				(void*)((unsigned
				long)src+i*SIZEOFWORD_BYTES),
				(void*)((unsigned
				long)dst+i*SIZEOFWORD_BYTES),
				word);
#endif
	}
	return 0;
}
/***********************************************************************
 * activate_fltd
 ***********************************************************************
 * Input:     fi_fltd_t *fltd  =  fault description
 * Output:    none
 * Effects:   takes actions necessary to activate a fault (including a fault of
 *            type "duration end", which actually removes a fault from the list)
 * Called By: fi_tick in fi.c
***********************************************************************/
void
activate_fltd(fi_fltd_t *fltd)
{
	switch(fltd->type){
		case FAULT_PERMANENT:
			/* non-trans. faults are appended to act. fault list */
			send_frontend("ACTIVATE permanent fault: %d\n", fltd->id);
			if(fltd->location == FAULT_POWER) { /* external fault */
				inject_power() ;
			} else { /* peripheral fault */
				fltd_aappend(&fi_afdl_first,fltd);
			}
			break;
		case FAULT_TRANSIENT:
			/* transient faults are injected right away */
			send_frontend("ACTIVATE transient fault: %d\n", fltd->id);
			fi_inject_trans_fault(fltd);
			break;
		case FAULT_DURATION_END:
			/* these are not really faults but just a marker that
			 * the duration of a fault has ended and it should
			 * therefore be removed from the active fault list
			 */
			send_frontend("ACTIVATE duration end: %d\n", fltd->id);
			fltd_aremove(&fi_afdl_first, fltd->id);
			break;
		case FAULT_DURATION:
		default:
			/* illegal or new fault type */
			assert(0);
			break;
	}
}
/***********************************************************************
 * fi_in_range
 ***********************************************************************
 * Input:     loff_t  offset  =  offset (in bytes) to be checked
 *            u. long cnt  =  number (of bytes) to be read
 *            fi_fltd_t *fltd  =  pointer to fault description
 * Output:    0 (False), 1 (True)
 * Effect:    Returns True (1) if [offset:offset+cnt] intersects with
 *            [block:block+range], FALSE (0) otherwise.
 *            blocknum is always in range if we have a complete hard disk
 *            failure (FAULT_HD_FAIL).
 * Called By: fi_watch_read_enter, fi_watch_write_enter
 **********************************************************************/
int
fi_in_range(
	loff_t offset,
	unsigned long cnt,
	fi_fltd_t *fltd)
{
	loff_t block, range ;
	assert(fltd != NULL) ;

	block =(fltd->fltd).hd.block ;
	range =(fltd->fltd).hd.range ;

	if (fltd->location == FAULT_HD_FAIL) {
		return 1;
	}
	if (fltd->location != FAULT_HD_BLOCKS) {
		return 0 ; /* should not happen */
	}
	if ( /* [offset:offset+cnt] and [block:block+range] don't overlap */
	     /* [offset:offset+cnt] not included in [block:block+range] */
		(    !( block <= offset     && offset     <= block+range )
		  && !( block <= offset+cnt && offset+cnt <= block+range )
		)
		&&
	     /* [block:block+range] not included in [offset:offset+cnt] */
		!(       offset <= block    && block+range <= offset+cnt )
		){
		return 0;
	} else {
		return 1;
	}
}
/***********************************************************************
 * functions for watching real mode system calls
 * they are needed to inject peripheral faults
***********************************************************************/
/***********************************************************************
 * fi_annull_syscall_enter
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effects:   Replaces system call (= virtual hardware access) with getpid().
 *            The number of the original syscall is saved in fi_.syscall.
 * Called By: fi_watch_{read,write,sendto}_enter
***********************************************************************/
void
fi_annull_syscall_enter(struct user_regs_struct *regs)
{
	int ret;

	FI_SC_SAVE_ORIG_FROM(regs->orig_eax) ;
	regs->orig_eax = __NR_getpid ;
	regs->eax = __NR_getpid;
	ret = ptrace(PTRACE_SETREGS, fi_.pid, 0, (unsigned long)regs);
	assert(0 <= ret);
	send_frontend("    annulled syscall %ld\n",fi_.syscall);
}
/***********************************************************************
 * fi_annull_syscall_return
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effects:   Restore regs->orig_eax to original value if this system call was
 *            annulled on entry.
 * Called By: fi_watch_return
***********************************************************************/
void
fi_annull_syscall_return(struct user_regs_struct *regs)
{
	assert(FI_SC_ANNULLED_ON_ENTRY);
	FI_SC_RESTORE_ORIG_TO(regs->orig_eax) ;

}
/***********************************************************************
 * fi_watch_open_enter
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    On a (real) open() system call, checks its parameters and compares
 *            them to the filenames of the currently active hd-faults. For
 *            open(), parameters are 
 *            char *pathname  ->  ebx
 *            int flags       ->  ecx
 *            mode_t mode     ->  edx.
 *            When a match is found, a pointer to the first matching fault
 *            description  is saved in fi_.watch_open_ptr.
 *            On return from the open() call the corresponding file descriptor
 *            (if any) is saved.
 * Called By: fi_watch_enter
 **********************************************************************/
#if 0
void
fi_watch_open_enter(struct user_regs_struct* regs)
{
	char *pathname = (char *)regs->ebx;

	fi_fltd_t *current = fi_afdl_first;

	assert(fi_.watch_enter & FI_WATCH_OPEN);
	fi_.watch_enter  &= ~FI_WATCH_OPEN;
	fi_.watch_return |=  FI_WATCH_OPEN;
	fi_.watch_enter  |= FI_WATCH_LSEEK;
	fi_.watch_enter  |= FI_WATCH_READ;
	fi_.watch_enter  |= FI_WATCH_WRITE;
	fi_.watch_enter  |= FI_WATCH_CLOSE;

	while (current != 0) {
		if ((current->location == FAULT_HD_BLOCKS)
				|| (current->location == FAULT_HD_FAIL)){
			if(!strncmp(pathname, (current->fltd).hd.device, FI_MAX_STRLEN)) {
				fi_.watch_open_ptr = current;
				break;
			}
		}
		current = current->next;
	}
	send_frontend("%s(%s) wop=0x%lx\n",  __FUNCTION__, pathname,
			(unsigned long)fi_.watch_open_ptr);
}
#endif
/***********************************************************************
 * fi_watch_open_return
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    On successful return from a (real mode) open() call, a
 *            filedescriptor is in eax. We then walk through the list of
 *            currently active fault descriptions and check (starting from first
 *            possible match as found in fi_watch_open_enter()) for hard disk
 *            faults affecting the current device. If we find any, we save
 *            pointers to the fault descriptions in (fi_.fds[fd]).a[]. On a
 *            complete hard disk failure, we stop looking and enter it as only
 *            pointer in above array.
 * Called By: fi_watch_return
***********************************************************************/
#if 0
void
fi_watch_open_return(struct user_regs_struct* regs)
{
	int fd = regs->eax;

	fi_fltd_t *current = fi_.watch_open_ptr ;/* may be NULL */

	assert(fi_.watch_return & FI_WATCH_OPEN);
	fi_.watch_enter  |=  FI_WATCH_OPEN;
	fi_.watch_return &= ~FI_WATCH_OPEN;

	if (0<=fd) {
		int i = 0;
		while (current != NULL) {
			if ((current->location == FAULT_HD_BLOCKS)
				|| (current->location == FAULT_HD_FAIL)){
				if(!strncmp((fi_.watch_open_ptr->fltd).hd.device,
					(current->fltd).hd.device, FI_MAX_STRLEN)) {
					if  (current->location == FAULT_HD_FAIL){
						(fi_.fds[fd]).a[0] = current;
						i=1;
						break;
					}
					(fi_.fds[fd]).a[i] = current;
					i++;
				}
			}
			current = current->next;
		}
		(fi_.fds[fd]).a[i] = NULL ; /* end of list marker */
		(fi_.fds[fd]).offset = 0;
		(fi_.fds[fd]).cnt = 0;
		(fi_.fds[fd]).buf = 0;
		(fi_.fds[fd]).open = 1;
		(fi_.fds[fd]).fault = 0;
	}
	fi_.watch_open_ptr = NULL;
	send_frontend("%d=%s() wop=0x%lx\n", fd, __FUNCTION__,
			(unsigned long)fi_.watch_open_ptr);
}
#endif
/***********************************************************************
 * fi_watch_lseek_enter
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    If lseek() is for one of the file descriptors listed in fi_.fds[],
 *            save the offset used in lseek(). Save the file-descriptor used in
 *            lseek() so fi_watch_lseek_return() knows what is going on.
 *            Parameters of lseek() are
 *            int fd        ->  ebx
 *            off_t offset  ->  ecx
 *            int whence    ->  edx
 * Called By: fi_watch_enter
 **********************************************************************/
void
fi_watch_lseek_enter(struct user_regs_struct* regs)
{
	int fd    = regs->ebx;
	loff_t offset  = (loff_t)(regs->ecx);

	assert(fi_.watch_enter & FI_WATCH_LSEEK);
	fi_.watch_enter  &= ~FI_WATCH_LSEEK;
	fi_.watch_return |=  FI_WATCH_LSEEK;
	fi_.watch_enter  |= FI_WATCH_READ;
	fi_.watch_enter  |= FI_WATCH_WRITE;

	send_frontend("%s(%d, 0x%llx)\n",
		__FUNCTION__, fd, offset) ;

	if (0<=fd) {
		if((fi_.fds[fd]).a[0] != NULL){
			/* at least one fault affects this operation */
			send_frontend("%s:affected\n", __FUNCTION__) ;
			(fi_.fds[fd]).offset = offset;
		}
		fi_.fd = fd;
	} else {
		fi_.fd = -1; /* this should not happen */
		assert(0) ;
	}
}
/***********************************************************************
 * fi_watch_llseek_enter
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    If llseek() is for one of the file descriptors listed in fi_.fds[],
 *            save the offset used in llseek(). Save the file-descriptor used in
 *            llseek() so fi_watch_llseek_return() knows what is going on.
 *            Parameters of lseek() are
 *            int fd                     ->  ebx
 *            unsigned long offset-high  ->  ecx
 *            unsigned long offset-low   ->  edx
 *	      loff_t *result             ->  esi
 *            unsigned int whence        ->  edi
 * Called By: fi_watch_enter
 **********************************************************************/
void
fi_watch_llseek_enter(struct user_regs_struct* regs)
{
	int fd    = regs->ebx;
	unsigned long offset_high  = (unsigned long)(regs->ecx);
	unsigned long offset_low   = (unsigned long)(regs->edx);
	loff_t offset = ((loff_t)offset_high<<32) | offset_low;

	assert(fi_.watch_enter & FI_WATCH_LSEEK);
	fi_.watch_enter  &= ~FI_WATCH_LSEEK;
	fi_.watch_return |=  FI_WATCH_LSEEK;
	fi_.watch_enter  |= FI_WATCH_READ;
	fi_.watch_enter  |= FI_WATCH_WRITE;

	send_frontend("%s(%d, 0x%llx)\n",
		__FUNCTION__, fd, offset);

	if (0<=fd) {
		if((fi_.fds[fd]).a[0] != NULL){
			/* at least one fault affects this operation */
			send_frontend("%s:affected\n", __FUNCTION__) ;
			(fi_.fds[fd]).offset = offset;
		}
		fi_.fd = fd;
	} else {
		fi_.fd = -1; /* this should not happen */
		assert(0) ;
	}
}
/***********************************************************************
 * fi_watch_lseek_return
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    If lseek() is for one of the file descriptors listed in fi_.fds[],
 *            save the offset and file-descriptor used in lseek().
 *            Return value of lseek() is offset in bytes from start of file
 *            returned in eax.
 * Called By: fi_watch_return
 **********************************************************************/
void
fi_watch_lseek_return(struct user_regs_struct* regs)
{
	loff_t offset    = (loff_t)regs->eax;

	assert(fi_.watch_return & FI_WATCH_LSEEK);
	fi_.watch_enter  |=  FI_WATCH_LSEEK;
	fi_.watch_return &= ~FI_WATCH_LSEEK;

	if (0<=fi_.fd) {
		if((fi_.fds[fi_.fd]).a[0] != NULL){
			/* at least one fault affects this operation */
			if ((fi_.fds[fi_.fd]).offset != offset) {
				/* this should not happen, right? */
				/* (fi_.fds[fi_.fd]).offset = 0; */
				/* assert(0) ; */
			}
		}
	}
	send_frontend("0x%llx=%s(%d, 0x%llx)\n",
		offset, __FUNCTION__,
		fi_.fd, (fi_.fds[fi_.fd]).offset) ;
}
void
fi_watch_llseek_return(struct user_regs_struct* regs)
{
	int ret ;
	loff_t offset ;

	ret = fi_peek_data(fi_.pid, (void *)regs->esi,
			&offset, sizeof(offset));
	assert(ret == 0);

	assert(fi_.watch_return & FI_WATCH_LSEEK);
	fi_.watch_enter  |=  FI_WATCH_LSEEK;
	fi_.watch_return &= ~FI_WATCH_LSEEK;

	if (0<=fi_.fd) {
		if((fi_.fds[fi_.fd]).a[0] != NULL){
			/* at least one fault affects this operation */
			if ((fi_.fds[fi_.fd]).offset != offset) {
				/* this should not happen, right? */
				/* (fi_.fds[fi_.fd]).offset = 0; */
				/* assert(0) ; */
			}
		}
	}
	send_frontend("0x%llx=%s(%d, 0x%llx)\n",
		offset,
		__FUNCTION__, fi_.fd, (fi_.fds[fi_.fd]).offset);
}
/***********************************************************************
 * fi_watch_read_enter
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    If read() is for one of the file descriptors listed in fi_.fds[],
 *            save the buffer and cnt used in read() so that
 *            fi_watch_read_return() knows what is going on. If this read() is
 *            indeed affected by a fault, annull it and set flag to let
 *            fi_watch_read_return() know.
 *            Parameters of read() are
 *            int fd       ->  ebx
 *            void *buffer ->  ecx
 *            int cnt      ->  edx
 * Called By: fi_watch_enter
 **********************************************************************/
void
fi_watch_read_enter(struct user_regs_struct* regs)
{
	int fd       = regs->ebx;
	void *buffer = (void *)(regs->ecx);
	unsigned long cnt      = regs->edx;

	if(fd == fi_.console.fd) { return ; }

	assert(fi_.watch_enter & FI_WATCH_READ);
	fi_.watch_enter  &= ~FI_WATCH_READ;
	fi_.watch_enter  &= ~FI_WATCH_WRITE;
	fi_.watch_return |=  FI_WATCH_READ;

	send_frontend("%s(%d, 0x%lx, %d)\n", __FUNCTION__,
		fd,(unsigned long)buffer,cnt);

	if ((0 <= fd) && (fi_.fds[fd]).a[0]!=NULL) {
		/* at least one fault affects this operation */
		int i = 0;
		loff_t offset = (fi_.fds[fi_.fd]).offset;
		assert(0 < cnt);
		fi_.fd = fd;
		(fi_.fds[fd]).buf = buffer;
		(fi_.fds[fd]).cnt = cnt;
		send_frontend("%s: affected %ld in %ld:%ld?\n",
				__FUNCTION__, offset,
		((fi_fltd_t*)((fi_.fds[fi_.fd]).a[i]))->fltd.hd.block,
		((fi_fltd_t*)((fi_.fds[fi_.fd]).a[i]))->fltd.hd.block
		+((fi_fltd_t*)((fi_.fds[fi_.fd]).a[i]))->fltd.hd.range);
		while ((fi_.fds[fi_.fd]).a[i] != NULL){
			if (fi_in_range(offset,cnt,
				(fi_fltd_t*)(fi_.fds[fi_.fd]).a[i])){
				send_frontend("%s: inject %lld:%lld in %ld:%ld!\n",
						__FUNCTION__, offset,
						(loff_t)(offset+cnt),
				((fi_fltd_t*)((fi_.fds[fi_.fd]).a[i]))->fltd.hd.block,
				((fi_fltd_t*)((fi_.fds[fi_.fd]).a[i]))->fltd.hd.block
				+((fi_fltd_t*)((fi_.fds[fi_.fd]).a[i]))->fltd.hd.range);
				/* inject fault */
				fi_annull_syscall_enter(regs);
				(fi_.fds[fd]).fault = 1;
				break;
			}
			i++;
		}
	}else{
		fi_.fd = -1;
	}
}
/***********************************************************************
 * fi_watch_read_return
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    If fi_watch_read_enter() annulled the read(), return -EIO. Reset
 *            some flags.
 * Called By: fi_watch_return
 **********************************************************************/
void
fi_watch_read_return(struct user_regs_struct* regs)
{
	int ret;
	int cnt;

	assert(fi_.watch_return & FI_WATCH_READ);
	/*
	fi_.watch_enter  |=  FI_WATCH_READ;
	*/
	fi_.watch_return &= ~FI_WATCH_READ;

	if (0 <= fi_.fd && (fi_.fds[fi_.fd]).a[0]!=NULL) {
		if ((fi_.fds[fi_.fd]).fault == 1) {
			/* inject fault */
			regs->eax = -EIO;
			ret = ptrace(PTRACE_SETREGS, fi_.pid, 0,
					(unsigned long)regs);
			assert(0 <= ret);
			send_frontend("%s:inject\n",__FUNCTION__);
		}
		cnt = regs->eax;
		if (0 < cnt){
			(fi_.fds[fi_.fd]).offset += cnt;
			(fi_.fds[fi_.fd]).cnt = cnt;
		}
	}
	(fi_.fds[fi_.fd]).fault = 0;
	send_frontend("%ld=%s()\n", (unsigned long)regs->eax, __FUNCTION__);
}
/***********************************************************************
 * fi_watch_write_enter
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    If write() is for one of the file descriptors listed in fi_.fds[],
 *            save the buffer and cnt used in write() so that
 *            fi_watch_write_return() knows what is going on. Anull write() if
 *            necessary and set flag.
 *            Parameters of write() are
 *            int fd       ->  ebx
 *            void *buffer ->  ecx
 *            int cnt      ->  edx
 * Called By: fi_watch_enter
 **********************************************************************/
void
fi_watch_write_enter(struct user_regs_struct* regs)
{
	int fd       = regs->ebx;
	void *buffer = (void *)(regs->ecx);
	unsigned long cnt      = regs->edx;

	if(fd == fi_.console.fd) { return ; }

	assert(fi_.watch_enter & FI_WATCH_WRITE);
	fi_.watch_enter  &= ~FI_WATCH_WRITE;
	fi_.watch_enter  &= ~FI_WATCH_READ;
	fi_.watch_return |=  FI_WATCH_WRITE;

	send_frontend("%s(%d, 0x%lx, %d)\n", __FUNCTION__,
			fd,(unsigned long)buffer,cnt);

	if ((0 <= fd) && ((fi_.fds[fd]).a[0] != NULL)) {
		loff_t offset = (fi_.fds[fi_.fd]).offset ;
		int i = 0;
		fi_.fd = fd;
		assert(0 < cnt);
		(fi_.fds[fd]).buf = buffer;
		(fi_.fds[fd]).cnt = cnt;
		send_frontend("%s: affected %lld:%lld in %ld:%ld?\n",
				__FUNCTION__, 
				offset,(loff_t)(offset+cnt),
		((fi_fltd_t*)((fi_.fds[fi_.fd]).a[i]))->fltd.hd.block,
		((fi_fltd_t*)((fi_.fds[fi_.fd]).a[i]))->fltd.hd.block
		+((fi_fltd_t*)((fi_.fds[fi_.fd]).a[i]))->fltd.hd.range);
		while ((fi_.fds[fi_.fd]).a[i] != NULL){
			if(fi_in_range(offset,cnt,
				(fi_fltd_t*)(fi_.fds[fi_.fd]).a[i])){
				send_frontend("%s: inject %lld:%lld in %ld:%ld\n",
						__FUNCTION__,
						offset,
						(loff_t)(offset+cnt),
				((fi_fltd_t*)((fi_.fds[fi_.fd]).a[i]))->fltd.hd.block,
				((fi_fltd_t*)((fi_.fds[fi_.fd]).a[i]))->fltd.hd.block
				+((fi_fltd_t*)((fi_.fds[fi_.fd]).a[i]))->fltd.hd.range);
				/* inject fault */
				fi_annull_syscall_enter(regs);
				(fi_.fds[fd]).fault = 1;
				break;
			}
			i++;
		}
	}else{
		fi_.fd = -1;
	}
}
/***********************************************************************
 * fi_watch_write_return
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    If fi_watch_write_enter() annulled write, return -EIO and reset
 *            flags.
 * Called By: fi_watch_return
 **********************************************************************/
void
fi_watch_write_return(struct user_regs_struct* regs)
{
	int cnt;
	int ret;

	assert(fi_.watch_return & FI_WATCH_WRITE);
	/*
	fi_.watch_enter  |=  FI_WATCH_WRITE;
	*/
	fi_.watch_return &= ~FI_WATCH_WRITE;
	if (0 <= fi_.fd && (fi_.fds[fi_.fd]).a[0]!=NULL) {
		if ((fi_.fds[fi_.fd]).fault == 1) {
			/* inject fault */
			regs->eax = -EIO;
			ret = ptrace(PTRACE_SETREGS, fi_.pid, 0,
					(unsigned long)regs);
			assert(0 <= ret);
		}
		cnt = regs->eax;
		if (0 < cnt) {
			(fi_.fds[fi_.fd]).offset += cnt;
			(fi_.fds[fi_.fd]).cnt = cnt;
		}
		(fi_.fds[fi_.fd]).fault = 0;
	}
	send_frontend("%ld=%s()\n",
			(unsigned long)regs->eax, __FUNCTION__);
}
/***********************************************************************
 * fi_watch_close_enter
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    Save file-descriptor used by close() for fi_watch_close_return().
 * Called By: fi_watch_enter
 **********************************************************************/
#if 0
void
fi_watch_close_enter(struct user_regs_struct* regs)
{
	int fd = regs->ebx;

	send_frontend("%s(%ld)\n", __FUNCTION__, fd) ;
	assert(fi_.watch_enter & FI_WATCH_CLOSE);
	fi_.watch_enter  &= ~FI_WATCH_CLOSE;
	fi_.watch_return |=  FI_WATCH_CLOSE;

	if (0 <= fd) {
		fi_.fd = fd;
	}else{
		fi_.fd = -1;
	}
}
#endif
/***********************************************************************
 * fi_watch_close_return
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    If read() is for one of the file descriptors listed in fi_.fds[],
 *            delete it from the list.
 * Called By: fi_watch_return
 **********************************************************************/
#if 0
void
fi_watch_close_return(struct user_regs_struct* regs)
{

	assert(fi_.watch_return & FI_WATCH_CLOSE);
	fi_.watch_enter  |=  FI_WATCH_CLOSE;
	fi_.watch_return &= ~FI_WATCH_CLOSE;
	fi_.watch_enter &= ~FI_WATCH_LSEEK;
	fi_.watch_enter &= ~FI_WATCH_READ;
	fi_.watch_enter &= ~FI_WATCH_WRITE;

	if (0 <= fi_.fd) {
	/* FIXME: check range */
		(fi_.fds[fi_.fd]).a[0] = NULL;
		(fi_.fds[fi_.fd]).open = 0;
	}
	send_frontend("%ld=%s()\n",
			(unsigned long)regs->eax, __FUNCTION__);
}
#endif
/***********************************************************************
 * functions used for network fault injection
***********************************************************************/
/***********************************************************************
 * fi_watch_bind_enter
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    bind() is called only once during bootup for each network device.
 *            Later calls to sendto() and recvfrom() only use the
 *            socket-descriptor returned by bind(). We must therefore use the
 *            bind() call to associate the port of the socket (known a priori)
 *            with the socket-descriptor. We have a list of 
 *            device-number : port number
 *            in fi_.netinfo[]. Now we want to extend the list to
 *            device-number : port number : socket-descriptor.
 *            If a flag is set, (fi_.netinfo[].fd == -2), we must also update
 *            the (already active) fault description.
 *            Parameters of bind() are
 *            SYS_BIND     ->  ebx
 *            void *args   ->  ecx
 * Called By: fi_watch_socketcall_enter
 **********************************************************************/
#if 0
void
fi_watch_bind_enter(struct user_regs_struct* regs)
{
	/* from src/include/asm/lotol.h, definition of lotol_bind */
	struct args_s{
		int fd;
		struct sockaddr *addr;
		int addrlen;
	} args;
	int  ret;
	struct sockaddr saddr;
	fi_fltd_t *fdl_current = NULL;
	unsigned short port;
	int i=0 ;

	fi_dprintf("%s(%d, %p)\n", __FUNCTION__, regs->ebx, regs->ecx);

	assert(regs->ebx == SYS_BIND);
	assert(regs->ecx != 0);

	ret = fi_peek_data(fi_.pid, (void *) regs->ecx,
			&args, sizeof(struct args_s));
	assert(ret == 0);
	assert(args.addrlen == sizeof(struct sockaddr));

	ret = fi_peek_data(fi_.pid, (void *) args.addr,
			&saddr, args.addrlen);
	assert(ret == 0);

	assert(((struct sockaddr_in*) &saddr)->sin_family == PF_INET);
	port = ntohs(((struct sockaddr_in*) &saddr)->sin_port);

	/*
	fi_dprintf("%s: port=0x%x=%d, fd=%d\n", __FUNCTION__, port, port, args.fd);
	*/

	/* Now update netinfo with the file descriptor belonging to the socket using
	 * the given port.
	 */
	for (i=0 ; i<MAX_ETH_CARDS ; i++) {
		/*
		fi_dprintf("%s: 0x%x == 0x%x ?", __FUNCTION__,
			fi_.netinfo[i].port, port) ;
		*/
		if (fi_.netinfo[i].port == port) {
			/*
			fi_dprintf(" yes\n") ;
			fi_dprintf("  %s.-2 == %d ?", __FUNCTION__, 
				fi_.netinfo[i].fd) ;
			*/
			if(-2 == fi_.netinfo[i].fd){
				/*
				fi_dprintf(" yes\n") ;
				*/
			/* we already have an active fault for this interface, so update
			 * fault description and the list of filedescriptors used for
			 * accessing devices with faults.
			 */
				for (fdl_current = fi_afdl_first;
					fdl_current != NULL;
					fdl_current = fdl_current->next) {
					/*
					fi_dprintf("    ((%d == .%d) || (%d == .%d)) && (0x%x == 0x%x)", fdl_current->location , FAULT_NET_RECV, fdl_current->location , FAULT_NET_SEND, fdl_current->fltd.net.port , port ) ;
					*/
					if ((fdl_current->location==FAULT_NET_RECV
						|| fdl_current->location==FAULT_NET_SEND)
						&& fdl_current->fltd.net.port==port) {
						/*
						fi_dprintf(" yes, fd=%d\n",args.fd) ;
						*/
						fdl_current->fltd.net.fd=args.fd ;
						/* select index into a: 0=SEND, 1=RECV */
						ret = ((fdl_current->location == FAULT_NET_SEND)?0:1) ;
						fi_.fds[args.fd].a[ret] = fdl_current;
						/* continue to look, we might
						 * already have more than one
						 * fault for this interface.
						 */
					} else {
						/*
						fi_dprintf(" no\n") ;
						*/
					}
				}
			} else {
				/*
				fi_dprintf(" no\n") ;
				*/
			}
			fi_.netinfo[i].fd = args.fd;
			break ;
		} else {
			/*
			fi_dprintf(" no\n") ;
			*/
		}
	}
	fi_.netinit-- ; /* decrease number of non-initialized nics */
}
#endif
/***********************************************************************
 * fi_watch_bind_return
 ***********************************************************************
 * There is no need for this function.
 **********************************************************************/

/***********************************************************************
 * fi_watch_connect_enter
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    connect() is called only once during bootup for the console
 *            Parameters of connect() are
 *            SYS_CONNECT  ->  ebx
 *            void *args   ->  ecx
 * Called By: fi_watch_socketcall_enter
 **********************************************************************/
#if 0
void
fi_watch_connect_enter(struct user_regs_struct* regs)
{
	/* from src/include/asm/lotol.h, definition of lotol_bind */
	struct args_s{
		int fd;
		struct sockaddr *addr;
		int addrlen;
	} args;
	int  ret;
	struct sockaddr saddr;
	unsigned short port;

	fi_dprintf("%s(%d, %p)\n", __FUNCTION__, regs->ebx, regs->ecx);

	assert(regs->ebx == SYS_CONNECT);
	assert(regs->ecx != 0);

	ret = fi_peek_data(fi_.pid, (void *) regs->ecx,
			&args, sizeof(struct args_s));
	assert(ret == 0);
	assert(args.addrlen == sizeof(struct sockaddr));

	ret = fi_peek_data(fi_.pid, (void *) args.addr,
			&saddr, args.addrlen);
	assert(ret == 0);

	assert(((struct sockaddr_in*) &saddr)->sin_family == PF_INET);
	port = ntohs(((struct sockaddr_in*) &saddr)->sin_port);

	if (port == fi_.console.port) {
		fi_.console.fd = args.fd ;
		fi_.watch_enter   &= ~FI_WATCH_CONNECT ;
		fi_dprintf("console: fd=%d, port=0x%hx\n", args.fd, port) ;
	}
}
#endif
/***********************************************************************
 * fi_watch_connect_return
 ***********************************************************************
 * There is no need for this function.
 **********************************************************************/

/***********************************************************************
 * fi_watch_sendto_enter
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    If sendto() is for one of the file descriptors listed in
 *            fi_.fds[], annull sendto() and set flag, so that
 *            fi_watch_sendto_return() knows to fake the return value of
 *            sendto() (really of getpid()) to return an error.
 * Called By: fi_watch_socketcall_enter
 **********************************************************************/
void
fi_watch_sendto_enter(struct user_regs_struct* regs)
{
	struct args_s {
		int s;
		void *buf;
		int len;
		int flags;
		struct lotol_sockaddr *to;
		int tolen;
	} args;
	int ret, fd;


	assert(regs->ebx == SYS_SENDTO);
	assert(regs->ecx != 0);

	assert(fi_.watch_enter & FI_WATCH_SENDTO);
	fi_.watch_enter  &= ~FI_WATCH_SENDTO;
	fi_.watch_return |=  FI_WATCH_SENDTO;

	ret = fi_peek_data(fi_.pid, (void *)regs->ecx,
			&args, sizeof(struct args_s));
	assert(ret == 0);
	fd = args.s;

	send_frontend("%s(%d, %p) %d\n", __FUNCTION__,
			regs->ebx, regs->ecx, fd);
	assert(0<=fd && fd < FI_MAX_FLTD);
	if ((fi_.fds[fd].a[0] != NULL) &&
		(((fi_fltd_t*)(fi_.fds[fd].a[0]))->location==FAULT_NET_SEND)){
		fi_annull_syscall_enter(regs);
		(fi_.fds[fd]).fault = 1;
		send_frontend("%s: inject\n", __FUNCTION__) ;
	}
	fi_.fd = fd;
}
/***********************************************************************
 * fi_watch_sendto_return
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    If sendto() was for one of the file descriptors listed in
 *            fi_.fds[], check flag set by fi_watch_sendto_enter() and
 *            fake the return value to return an -EIO in case it is set.
 * Called By: fi_watch_socketcall_return
 **********************************************************************/
void
fi_watch_sendto_return(struct user_regs_struct* regs)
{
	int ret;
	assert(fi_.watch_return & FI_WATCH_SENDTO);
	fi_.watch_enter  |=  FI_WATCH_SENDTO;
	fi_.watch_return &= ~FI_WATCH_SENDTO;

	assert(0<=fi_.fd && fi_.fd< FI_MAX_FLTD);
	if ((fi_.fds[fi_.fd].a[0] != NULL)
		&& fi_.fds[fi_.fd].fault) {
		/* if this system call was annulled by watch_*_enter */
		regs->eax = -EIO;
		ret = ptrace(PTRACE_SETREGS, fi_.pid, 0,
				(unsigned long)regs);
		assert(0 <= ret);
		send_frontend("%s: inject\n", __FUNCTION__) ;
	}
	fi_.fds[fi_.fd].fault = 0;
	fi_.fd = -1;
	send_frontend("%ld=%s()\n",
			(unsigned long)regs->eax, __FUNCTION__);
}
/***********************************************************************
 * fi_watch_recvfrom_enter
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    If recvfrom() is for one of the file socket-escriptors listed in
 *            network fault descriptions, set a flag, so
 *            fi_watch_recvfrom_return() knows to fake the return value of
 *            recvfrom() to return an error.
 * Called By: fi_watch_socketcall_enter
 **********************************************************************/
void
fi_watch_recvfrom_enter(struct user_regs_struct* regs)
{
	struct args_s {
		int s;
		void *buf;
		int len;
		int flags;
		struct lotol_sockaddr *from;
		int *fromlen;
	} args;
	fi_fltd_t *current;
	int ret;
	int fd;

	assert(regs->ebx == SYS_RECVFROM);
	assert(regs->ecx != 0);

	assert(fi_.watch_enter & FI_WATCH_RECVFROM);

	fi_.watch_enter  &= ~FI_WATCH_RECVFROM;
	fi_.watch_return |=  FI_WATCH_RECVFROM;

	ret = fi_peek_data(fi_.pid, (void *)regs->ecx,
			&args, sizeof(struct args_s));
	assert(ret == 0);
	fd = args.s;

	assert(0 <= fd && fd < FI_MAX_FLTD);
	send_frontend("%s(%d, %p) %d\n", __FUNCTION__,
			regs->ebx, regs->ecx, fd);

	for (current = fi_afdl_first;
		current != NULL;
		current = current->next) {
		if (current->location == FAULT_NET_RECV
			&& current->fltd.net.fd == fd) {
			send_frontend("%s: inject\n", __FUNCTION__);
			fi_.fds[fd].fault = 1;
			break;
		}
	}
	fi_.fd = fd;
}
/***********************************************************************
 * fi_watch_recvfrom_return
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    If recvfrom() was for one of the socket-descriptors belonging to
 *            faulty devices, check flag set by fi_watch_recvfrom_enter() and
 *            fake the return value to return an -EIO in case it is set.
 * Called By: fi_watch_socketcall_return
 **********************************************************************/
void
fi_watch_recvfrom_return(struct user_regs_struct* regs)
{
	int ret;

	assert(fi_.watch_return & FI_WATCH_RECVFROM);

	fi_.watch_enter  |=  FI_WATCH_RECVFROM;
	fi_.watch_return &= ~FI_WATCH_RECVFROM;

	assert(0 <= fi_.fd && fi_.fd < FI_MAX_FLTD);

	if (fi_.fds[fi_.fd].fault) {

		if (0 < regs->eax) {
			regs->eax = -EIO;
			ret = ptrace(PTRACE_SETREGS, fi_.pid, 0, (unsigned long)regs);
			assert(0 <= ret);
		}
		send_frontend("%s: inject\n", __FUNCTION__) ;

		fi_.fds[fi_.fd].fault = 0;
	}
	fi_.fd = -1;

	send_frontend("%ld=%s()\n",
			(unsigned long)regs->eax, __FUNCTION__);
}
/***********************************************************************
 * fi_watch_socketcall_enter
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    Watches commmon kernel entry point to socketcalls and saves the
 *            number of the socketcall made for
 *            fi_watch_{bind,sendto,recvfrom}_enter.
 *            Parameters of socketcall() are
 *            int call     ->  ebx
 *            u. long args ->  ecx
 * Called By: fi_watch_enter
 **********************************************************************/
void
fi_watch_socketcall_enter(struct user_regs_struct* regs)
{
	fi_.socketcall = regs->ebx;
	switch (fi_.socketcall) {
		case SYS_SOCKET: /* 1 */
			break;
		case SYS_BIND: /* 2 */
			if(fi_.watch_enter & FI_WATCH_BIND){
				/*
				fi_watch_bind_enter(regs);
				*/
			}
			break;
		case SYS_CONNECT: /* 3 */
			if(fi_.watch_enter & FI_WATCH_CONNECT){
				/*
				fi_watch_connect_enter(regs);
				*/
			}
			break;
		case SYS_LISTEN: /* 4 */
		case SYS_ACCEPT: /* 5 */
		case SYS_GETSOCKNAME: /* 6 */
		case SYS_GETPEERNAME: /* 7 */
		case SYS_SOCKETPAIR: /* 8 */
		case SYS_SEND: /* 9 */
		case SYS_RECV: /* 10 */
			break;
		case SYS_SENDTO: /* 11 */
			if(fi_.watch_enter & FI_WATCH_SENDTO){
				fi_watch_sendto_enter(regs);
			}
			break;
		case SYS_RECVFROM: /* 12 */
			if(fi_.watch_enter & FI_WATCH_RECVFROM){
				fi_watch_recvfrom_enter(regs);
			}
			break;
		case SYS_SHUTDOWN: /*13 */
		case SYS_SETSOCKOPT: /* 14 */
		case SYS_GETSOCKOPT: /* 15 */
		case SYS_SENDMSG: /* 16 */
		case SYS_RECVMSG: /* 17 */
		default:
			break;
	}
}
/***********************************************************************
 * fi_watch_socketcall_enter
 ***********************************************************************
 * Input:     struct user_regs_struct *regs  =  current register contens of
 *            UMLinux virtual machine
 * Output:    none
 * Effect:    Dispatcher for fi_watch_{bind,sendto,recvfrom}_return. Number of
 *            socketcall still in fi_.socketcall.
 * Called By: fi_watch_return
 **********************************************************************/
void
fi_watch_socketcall_return(struct user_regs_struct* regs)
{
	switch (fi_.socketcall) {
		case SYS_SOCKET:
			break;
		case SYS_BIND:
			/* so far fi_watch_bind_return() not necessary */
			break;
		case SYS_SENDTO:
			fi_watch_sendto_return(regs);
			break;
		case SYS_RECVFROM:
			fi_watch_recvfrom_return(regs);
			break;
		default:
			break;
	}
}
/***********************************************************************
 * functions for injecting computing core faults
***********************************************************************/
/***********************************************************************
 * inject_ram
 ***********************************************************************
 * Input:     fi_fltd_t *fltd  =  fault description
 * Output:    none
 * Effect:    Read actual RAM content at specified address, flip specified bit
 *            and write it back to UMLinux virtual machine.
 * Called By: fi_inject_trans_fault
 **********************************************************************/
void
inject_ram(fi_fltd_t *fltd)
{
	long ret;

	send_frontend("%s(%p)\n", __FUNCTION__, fltd) ;
	assert(fltd != NULL);
	assert(fltd->location == FAULT_RAM);
	assert(0 <= fltd->fltd.ram.addr
	    && fltd->fltd.ram.addr < config.mem_size
	    && ! (fltd->fltd.ram.addr & 3));
	assert(0 <= fltd->fltd.ram.bit
	    && fltd->fltd.ram.bit < 32);

	/* read actual ram content */
	errno = 0;
	ret = ptrace(PTRACE_PEEKDATA, fi_.pid,
		(void *) (fltd->fltd.ram.addr + MEM_BASE), 0);
	send_frontend("contents at 0x%08lx: %08lx\n",
		fltd->fltd.ram.addr, ret);
	assert(errno == 0);

	/* flip bit */
	ret = flip_bit(fltd->fltd.ram.bit, ret) ;
	send_frontend("contents at 0x%08lx: %08lx\n",
		fltd->fltd.ram.addr, ret);

	/* write back word */
	ret = ptrace(PTRACE_POKEDATA, fi_.pid,
		(void *) (fltd->fltd.ram.addr + MEM_BASE), &ret);
	assert(0 <= ret);
}

/***********************************************************************
 * inject_cpu_regs
 ***********************************************************************
 * Input:     fi_fltd_t *fltd  =  fault description
 * Output:    none
 * Effect:    Read actual content of specified CPU-register, flip specified bit
 *            and write it back to UMLinux virtual machine.
 * Called By: fi_inject_trans_fault
 **********************************************************************/
void
inject_cpu_regs(fi_fltd_t *fltd)
{
	struct user_regs_struct regs;
	int ret;

	send_frontend("%s(%p)\n", __FUNCTION__, fltd) ;
	assert(fltd != NULL);
	assert(fltd->location == FAULT_CPU_REGS);

	ret = ptrace(PTRACE_GETREGS, fi_.pid, 0, (unsigned long) &regs);
	fi_dump_regs(&regs);
	send_reg_frontend(&regs, (fltd->fltd).cpu.reg);
	assert(0 <= ret);
	switch ((fltd->fltd).cpu.reg) {
	case CPU_REG_EAX:
		regs.eax = (long) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.eax)); 
		break;
	case CPU_REG_EBX:
		regs.ebx = (long) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.ebx)); 
		break;
	case CPU_REG_ECX:
		regs.ecx = (long) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.ecx)); 
		break;
	case CPU_REG_EDX:
		regs.edx = (long) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.edx)); 
		break;
	case CPU_REG_ESI:
		regs.esi = (long) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.esi)); 
		break;
	case CPU_REG_EDI:
		regs.edi = (long) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.edi)); 
		break;
	case CPU_REG_EBP:
		regs.ebp = (long) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.ebp)); 
		break;
	case CPU_REG_ESP:
		regs.esp = (long) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.esp)); 
		break;
	case CPU_REG_EIP:
		regs.eip = (long) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.eip)); 
		break;
	case CPU_REG_DS:
		regs.xds=(unsigned short) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.xds)); 
		break;
	case CPU_REG_CS:
		regs.xcs=(unsigned short) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.xcs)); 
		break;
	case CPU_REG_SS:
		regs.xss=(unsigned short) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.xss)); 
		break;
	case CPU_REG_ES:
		regs.xes=(unsigned short) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.xes)); 
		break;
	case CPU_REG_FS:
		regs.xfs=(unsigned short) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.xfs)); 
		break;
	case CPU_REG_GS:
		regs.xgs=(unsigned short) flip_bit((fltd->fltd).cpu.bit,
				(unsigned long) (regs.xgs)); 
		break;
	default:
		assert(0) ;
		break;
	}
	ret = ptrace(PTRACE_SETREGS, fi_.pid, 0, (unsigned long) &regs);
	fi_dump_regs(&regs);
	send_reg_frontend(&regs, (fltd->fltd).cpu.reg);
	assert(0 <= ret);
}
/***********************************************************************
 * fi_inject_trans_fault
 ***********************************************************************
 * Input:     fi_fltd_t *fltd  =  fault description
 * Output:    none
 * Effect:    Injects a transient fault.
 * Called By: activate_fltd
 **********************************************************************/
void
fi_inject_trans_fault(fi_fltd_t *fltd)
{
	assert(fltd != NULL);
	switch (fltd->location){
		case FAULT_RAM:
			inject_ram(fltd);
			break;
		case FAULT_CPU_REGS:
			inject_cpu_regs(fltd);
			break;
		default:
			send_frontend("illegal combination %s + %s\n",
				fltd->location, fltd->type);
			assert(0) ;
			break;
	}
}
/***********************************************************************
 * fi_watch_enter
 ***********************************************************************
 * Input:     int pid   =   process ID of UMLinux virtual machine 
 *            struct user_regs_struct *regs  =  current register contens of
 *                                              UMLinux virtual machine
 * Output:    none
 * Effect:    If peripheral faults are active, calls the appropriate
 *            fi_watch_*_enter() functions if necessary.
 * Called By: start_smsyscall (tracer.c)
 **********************************************************************/
void
fi_watch_enter(
	int pid,
	struct user_regs_struct*regs)
{
	int ret;
	if (fi_.ncpu != 1) { return; }

	if (PERIPHERAL_FAULTS_ACTIVE){
		ret = ptrace(PTRACE_GETREGS, pid, 0, (unsigned long) regs);
		assert(0 <= ret);
		if ((fi_.watch_enter & FI_WATCH_OPEN)
				&&(regs->orig_eax == __NR_open)){
			/*
			fi_watch_open_enter(regs);
			*/
		}
		if (fi_.watch_enter & FI_WATCH_LSEEK){
			if (regs->orig_eax == __NR_lseek) {
				fi_watch_lseek_enter(regs);
			} else if (regs->orig_eax == __NR__llseek) {
				fi_watch_llseek_enter(regs);
			}
		}
		if ((fi_.watch_enter & FI_WATCH_READ)
				&&(regs->orig_eax == __NR_read)) {
			fi_watch_read_enter(regs);
		}
		if ((fi_.watch_enter & FI_WATCH_WRITE)
				&&(regs->orig_eax == __NR_write)) {
			fi_watch_write_enter(regs);
		}
		if ((fi_.watch_enter & FI_WATCH_CLOSE)
				&&(regs->orig_eax == __NR_close)) {
			/*
			fi_watch_close_enter(regs);
			*/
		}
		if ((fi_.watch_enter & FI_WATCH_SOCKETCALL)
				&&(regs->orig_eax == __NR_socketcall)) {
			fi_watch_socketcall_enter(regs);
		}
	}
	if (fi_.netinit && regs->orig_eax == __NR_socketcall) {
		fi_watch_socketcall_enter(regs);
	}
}
/***********************************************************************
 * fi_watch_return
 ***********************************************************************
 * Input:     int pid   =   process ID of UMLinux virtual machine 
 *            struct user_regs_struct *regs  =  current register contens of
 *                                              UMLinux virtual machine
 * Output:    none
 * Effect:    If peripheral faults are active, calls the appropriate
 *            fi_watch_*_return() functions if necessary.
 * Called By: do_smsyscall (tracer.c)
 **********************************************************************/
void
fi_watch_return(
	int pid,
	struct user_regs_struct*regs)
{
	int ret;
	if (fi_.ncpu != 1) { return; }
	/*
	int fd ;

	fi_fltd_t *current ;
	*/

	if (PERIPHERAL_FAULTS_ACTIVE){
		ret = ptrace(PTRACE_GETREGS, pid, 0, (unsigned long) regs);
		assert(0 <= ret);
		if (FI_SC_ANNULLED_ON_ENTRY){
			fi_annull_syscall_return(regs);
		}
		if ((fi_.watch_return & FI_WATCH_OPEN)
				&&(regs->orig_eax == __NR_open)){
			/*
			fi_watch_open_return(regs);
			*/
		}
		if (fi_.watch_return & FI_WATCH_LSEEK){
			if (regs->orig_eax == __NR_lseek){
				fi_watch_lseek_return(regs);
			} else if (regs->orig_eax == __NR__llseek) {
				fi_watch_llseek_return(regs);
			}
		}
		if ((fi_.watch_return & FI_WATCH_READ)
				&&(regs->orig_eax == __NR_read)) {
			fi_watch_read_return(regs);
		}
		if ((fi_.watch_return & FI_WATCH_WRITE)
				&&(regs->orig_eax == __NR_write)) {
			fi_watch_write_return(regs);
		}
		if ((fi_.watch_return & FI_WATCH_CLOSE)
				&&(regs->orig_eax == __NR_close)) {
			/*
			fi_watch_close_return(regs);
			*/
		}
		if ((fi_.watch_return & FI_WATCH_SOCKETCALL)
				&& (regs->orig_eax == __NR_socketcall)) {
			fi_watch_socketcall_return(regs);
		}
	}
}
/***********************************************************************
 * functions for injecting external faults
***********************************************************************/
/***********************************************************************
 * inject_power
 ***********************************************************************
 * Input:     none
 * Output:    none
 * Effect:    Send SIGKILL to UMLinux virtual machine.
 * Called By: activate_fltd
 **********************************************************************/
void
inject_power()
{
/*
	int ret;

	fi_dprintf("%s(%p)\n", __FUNCTION__, fltd) ;
	ret = kill(fi_.pid, SIGTERM);
	assert(0 <= ret);
*/
}
