/* $Id: fi_parse.c,v 1.1.1.1 2002/08/16 17:29:27 root Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */
#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "fi_defs.h"
#include "config.h"
#include "lotol_fdesc.h"

/* fi_sbuf_t fi_sbuf; */
extern fi_flt_inj_t fi_; /* fi.c */
extern int fi_node ; /* fi.c */

/***********************************************************************
 * function prototypes
 ***********************************************************************
 * file internal functions 
 **********************************************************************/
/*
void *fi_stralloc(int);
void  fi_strshrink(char *);
void *fi_fltdalloc(void);
*/

int read_string (int, char**);
int str_to_num(char *);

int parse_hd_fail(int, fi_fltd_t *);
int parse_hd_blocks(int, fi_fltd_t *);
int parse_ram(int, fi_fltd_t *);
int parse_cpu_regs(int, fi_fltd_t *);
int parse_net(int, fi_fltd_t *);
void exit_on_error(int ) ;
/***********************************************************************
 * exported functions
 **********************************************************************/
int parse_fltd(int, fi_fltd_t **); /* to fi.c */
int read_decnum (int, loff_t*); /* to fi.c */

/***********************************************************************
 * imported functions
 **********************************************************************/
extern void fi_dprintf(char *, ...); /* fi_print.c */
extern void fi_eprintf(char *, ...); /* fi_print.c */
/***********************************************************************
 * exit_on_error
 ***********************************************************************
 * Input:     int error  =  error number from fi_defs.h
 * Output:    none
 * Effects:   sends UMLinux virtual machine process a SIGKILL and exits.
 * Called By: read_string, parse_fltd
***********************************************************************/
void
exit_on_error(int error)
{
	switch (error) {
	case FI_EMULT:
		fi_eprintf("ERROR: multiplicity error\n");
		break;

	case FI_EPARSE:
		fi_eprintf("ERROR: parsing error\n");
		break;
	case ENOMEM:
		fi_eprintf("ERROR: no more memory.\n");
		fi_eprintf("Try raising limits for number of faults etc.\n");
		break;
	default:
		fi_eprintf("unknown error\n");
		break;
	}
	fi_eprintf("\n%d: Killing child and exiting.\n\n", fi_node);
	kill(fi_.pid, SIGKILL);
	exit(error);
}
/***********************************************************************
 * read_string
 ***********************************************************************
 * Input:     int socket  =   socket descriptor for comm. with frontend
 *            char **string  =  parsed string will be returned herein, != NULL
 * Output:    ENOMEM if no memory could be allocated for string
 *            FI_EPARSE if no valid string was found at this position
 *            0 if all went ok
 * Effects:   reads string consisting of a-zA-Z0-9_/ (/ for path names)
 * Called By: parse_*-functions
***********************************************************************/
int
read_string(
	int socket, char **string)
{
	long ret;
	int cnt;
	char c;

	assert(string != NULL);

	*string = (char *) calloc(1, sizeof(char)*(FI_MAX_STRLEN+1));
	if (*string == NULL) {
		exit_on_error(ENOMEM);
	}
	assert(*string != NULL);

	cnt = 0;
	ret = read(socket,&c,1) ;
	if(ret<0) {
		fi_dprintf("%s: %s\n", __FUNCTION__, strerror(errno)) ;
	}
	assert(1==ret) ;

	while ((('0' <= c && c <= '9')
		 || ('A' <= c && c <= 'Z')
		 || ('a' <= c && c <= 'z')
		 || ('_' == c)
		 || ('/' == c))
		 && (cnt < FI_MAX_STRLEN)) {
	fi_dprintf(":%c", c) ; fflush(stderr) ;
		(*string)[cnt]=(char)c;
		cnt++;
		ret = read(socket,&c,1) ;
		assert(1==ret) ;
	fi_dprintf(":\n") ; fflush(stderr) ;
	}
	realloc(*string, (strlen(*string)+1));
	if (cnt>0) {
		return 0;
	}
	/* if we are here, some error occurred. */
	free(*string) ;
	*string = NULL;
	return FI_EPARSE;
}
/***********************************************************************
 * read_decnum
 ***********************************************************************
 * Input:     int socket  =   socket descriptor for comm. with frontend
 *            u. long *decnum  =  parsed number will be returned herein, != NULL
 * Output:    FI_EPARSE if no valid number was found at this position
 *            0 if all went ok
 * Effects:   
 * Called By: parse_*-functions, fi_tick in fi.c
***********************************************************************/
int
read_decnum(
	int socket,
	loff_t *decnum)
{
	long ret;
	int cnt = -1;
	loff_t i = 0;
	char c = '0';

	assert(decnum != 0);

	while ('0' <= c && c <= '9') {
		i *= 10;
		i += c - '0';
		cnt++;
		ret = read(socket,&c,1) ;
		assert(1==ret) ;
		fi_dprintf(":%c:\n", c) ; fflush(stderr) ;
	}
	if (cnt>0) {
		*decnum = i;
		return 0;
	}
	return FI_EPARSE;
}
/***********************************************************************
 * str_to_num
 ***********************************************************************
 * Input:     char *s  =  string to convert to number
 * Output:    positive integer if conversion was possible,
 *            FI_EPARSE otherwise
 * Effects:   uses constants from fi_defs.h
 * Called By: parse_*-functions
***********************************************************************/
int
str_to_num(char *s)
{
	if (s == (char*)0 ) {
		return FI_EPARSE;
	}
	if(!strncmp(s,FAULT_RAM_STR,FI_MAX_STRLEN)) return FAULT_RAM;
	if(!strncmp(s,FAULT_CPU_REGS_STR,FI_MAX_STRLEN)) return FAULT_CPU_REGS;
	if(!strncmp(s,FAULT_MMU_STR,FI_MAX_STRLEN)) return FAULT_MMU;
	if(!strncmp(s,FAULT_HD_BLOCKS_STR,FI_MAX_STRLEN))return FAULT_HD_BLOCKS;
	if(!strncmp(s,FAULT_HD_FAIL_STR,FI_MAX_STRLEN)) return FAULT_HD_FAIL;
	if(!strncmp(s,FAULT_NET_SEND_STR,FI_MAX_STRLEN)) return FAULT_NET_SEND;
	if(!strncmp(s,FAULT_NET_RECV_STR,FI_MAX_STRLEN)) return FAULT_NET_RECV;
	if(!strncmp(s,FAULT_POWER_STR,FI_MAX_STRLEN))    return FAULT_POWER;
	if(!strncmp(s,FAULT_PERMANENT_STR,FI_MAX_STRLEN))return FAULT_PERMANENT;
	if(!strncmp(s,FAULT_TRANSIENT_STR,FI_MAX_STRLEN))return FAULT_TRANSIENT;
	if(!strncmp(s,FAULT_DURATION_STR,FI_MAX_STRLEN)) return FAULT_DURATION;
	if(!strncmp(s,FAULT_DURATION_END_STR,FI_MAX_STRLEN))
		return FAULT_DURATION_END;

	if(!strncmp(s,CPU_REG_EAX_STR, FI_MAX_STRLEN)) return CPU_REG_EAX;
	if(!strncmp(s,CPU_REG_EBX_STR, FI_MAX_STRLEN)) return CPU_REG_EBX;
	if(!strncmp(s,CPU_REG_ECX_STR, FI_MAX_STRLEN)) return CPU_REG_ECX;
	if(!strncmp(s,CPU_REG_EDX_STR, FI_MAX_STRLEN)) return CPU_REG_EDX;
	if(!strncmp(s,CPU_REG_ESI_STR, FI_MAX_STRLEN)) return CPU_REG_ESI;
	if(!strncmp(s,CPU_REG_EDI_STR, FI_MAX_STRLEN)) return CPU_REG_EDI;
	if(!strncmp(s,CPU_REG_EBP_STR, FI_MAX_STRLEN)) return CPU_REG_EBP;
	if(!strncmp(s,CPU_REG_ESP_STR, FI_MAX_STRLEN)) return CPU_REG_ESP;
	if(!strncmp(s,CPU_REG_EIP_STR, FI_MAX_STRLEN)) return CPU_REG_EIP;
	if(!strncmp(s,CPU_REG_DS_STR,  FI_MAX_STRLEN)) return CPU_REG_DS;
	if(!strncmp(s,CPU_REG_CS_STR,  FI_MAX_STRLEN)) return CPU_REG_CS;
	if(!strncmp(s,CPU_REG_SS_STR,  FI_MAX_STRLEN)) return CPU_REG_SS;
	if(!strncmp(s,CPU_REG_ES_STR,  FI_MAX_STRLEN)) return CPU_REG_ES;
	if(!strncmp(s,CPU_REG_FS_STR,  FI_MAX_STRLEN)) return CPU_REG_FS;
	if(!strncmp(s,CPU_REG_GS_STR,  FI_MAX_STRLEN)) return CPU_REG_GS;

	if(!strncmp(s,END_OF_FAULTLOAD_STR,  FI_MAX_STRLEN))
		return END_OF_FAULTLOAD;
	/* should not be reached */
	return FI_EPARSE;
}
int
find_hd_fd(char *device)
{
	int ret;
	char drive = '0';
	int fd = -1;
	ret = sscanf(device, "hd%1c/media", &drive);
	if (1!=ret){
		ret = sscanf(device, "fd%1c/media", &drive);
		assert(1==ret);
		drive -= '0';
		assert (0 <= drive && drive < MAX_FDS);
		fd = FD_FD_MEDIA(drive);
	}else{
		assert(1==ret);
		drive -= 'a';
		assert(0 <= drive && drive < MAX_HDS);
		fd = FD_HD_MEDIA(drive);
	}
	assert(0<fd);
	return(fd);
}
/***********************************************************************
 * parse_hd_fail
 ***********************************************************************
 * Input:     int socket  =   socket descriptor for comm. with frontend
 *            fi_ftld_t *fltd_new  =  pointer to new fault description, !=NULL!
 * Output:    FI_EPARSE when a parse error occurred
 *            0 if all went ok
 * Effect:    fills fltd with parsed information
 * Called By: parse_fltd
 **********************************************************************/
int
parse_hd_fail(
	int socket,
	fi_fltd_t *fltd_new)
{
	int ret;
	char *s;

	ret = read_string(socket, &s);
	if (ret!=0) {
		return FI_EPARSE;
	}
	fltd_new->fltd.hd.device = s ; /* valid string found */
	fltd_new->fltd.hd.block = -1 ; /* unused in complete hd failure*/
	fltd_new->fltd.hd.range = -1 ; /* unused in complete hd failure*/
	fltd_new->fltd.hd.fd = find_hd_fd(fltd_new->fltd.hd.device);

	return(0);
}
/***********************************************************************
 * parse_hd_blocks
 ***********************************************************************
 * Input:     int socket  =   socket descriptor for comm. with frontend
 *            fi_ftld_t *fltd_new  =  pointer to new fault description, !=NULL!
 * Output:    FI_EPARSE when a parse error occurred
 *            0 if all went ok
 * Effect:    fills *fltd_new with parsed information
 * Called By: parse_fltd
 **********************************************************************/
int parse_hd_blocks(int socket, fi_fltd_t *fltd_new)
{
	int ret;
	char *s;
	loff_t num;

	ret = read_string(socket, &s);
	fi_dprintf("parse_hd_blocks: string\n");
	if (ret!=0) {
		return FI_EPARSE;
	}
	fi_dprintf("   --> ok\n");
	(fltd_new->fltd).hd.device = s ; /* valid string found */

	ret = read_decnum(socket, &num);
	fi_dprintf("parse_hd_blocks: num\n");
	if (ret!=0) {
		return FI_EPARSE;
	}
	fi_dprintf("   --> ok\n");
	(fltd_new->fltd).hd.block = num ; /* valid block number  found */
	
	ret = read_decnum(socket, &num);
	fi_dprintf("parse_hd_blocks: num\n");
	if (ret!=0) {
		return FI_EPARSE;
	}
	fi_dprintf("   --> ok\n");
	(fltd_new->fltd).hd.range = num ; /* valid number found */
	fltd_new->fltd.hd.fd = find_hd_fd(fltd_new->fltd.hd.device);
	return(0);
}
/***********************************************************************
 * parse_ram
 ***********************************************************************
 * Input:     int socket  =   socket descriptor for comm. with frontend
 *            fi_ftld_t *fltd_new  =  pointer to new fault description, !=NULL!
 * Return:    FI_EPARSE when a parse error occurred
 *            0 if all went ok
 * Effect:    fills *fltd_new with parsed information
 * Called By: parse_fltd
 **********************************************************************/
int
parse_ram(
	int socket,
	fi_fltd_t *fltd_new)
{
	long ret;
	loff_t num;

	ret = read_decnum(socket, &num);
	fi_dprintf("parse_ram: num\n");
	if (ret!=0) {
		return FI_EPARSE;
	}
	fi_dprintf("   --> ok\n");
	if (config.mem_size <= num /* smaller than available memory */
		|| (num & 3)) { /* word aligned */
		return FI_EPARSE;
	}
	(fltd_new->fltd).ram.addr = num ; /* valid address found */

	ret = read_decnum(socket, &num);
	fi_dprintf("parse_ram: num\n");
	if (ret!=0) {
		return FI_EPARSE;
	}
	if (32 <= num) {
		return FI_EPARSE;
	}
	fi_dprintf("   --> ok\n");
	(fltd_new->fltd).ram.bit = num ; /* valid bit number  found */
	
	return(0);
}
/***********************************************************************
 * parse_cpu_regs
 ***********************************************************************
 * Input:     int socket  =   socket descriptor for comm. with frontend
 *            fi_ftld_t *fltd_new  =  pointer to new fault description, !=NULL!
 * Return:    FI_EPARSE when a parse error occurred
 *            0 if all went ok
 * Effect:    fills *fltd_new with parsed information
 * Called By: parse_fltd
 **********************************************************************/
int
parse_cpu_regs(
	int socket,
	fi_fltd_t *fltd_new)
{
	long ret;
	char *s;
	loff_t num;

	ret = read_string(socket, &s);
	fi_dprintf("parse_cpu_regs: string\n");
	if (ret!=0) {
		return FI_EPARSE;
	}
	num = str_to_num(s);
	if ( ! ((long)num>0) ) {
		return FI_EPARSE;
	}
	fi_dprintf("   --> ok\n");
	(fltd_new->fltd).cpu.reg = (int)num ; /* valid register found */

	ret = read_decnum(socket, &num);
	fi_dprintf("parse_cpu_regs: num\n");
	if (ret!=0) {
		return FI_EPARSE;
	}
	fi_dprintf("   --> ok\n");
	(fltd_new->fltd).cpu.bit = num ; /* valid bit number  found */
	
	return(0);
}
/***********************************************************************
 * parse_net
 ***********************************************************************
 * Input:     int socket  =   socket descriptor for comm. with frontend
 *            fi_ftld_t *fltd_new  =  pointer to new fault description, !=NULL!
 * Return:    FI_EPARSE when a parse error occurred
 *            0 if all went ok
 * Effect:    fills *fltd_new with parsed information
 * Called By: parse_fltd
 **********************************************************************/
int
parse_net(
	int socket,
	fi_fltd_t *fltd_new)
{
	int ret;
	/*
	int i ;
	*/
	int device;
	char *s;
	loff_t num ;

	ret = read_string(socket, &s);
	fi_dprintf("parse_net: string\n");
	if (ret!=0) {
		return FI_EPARSE;
	}
	fi_dprintf("   --> ok\n");
	(fltd_new->fltd).net.device = s ; /* valid device found */

	ret = read_decnum(socket, &num);
	fi_dprintf("parse_net: num\n");
	if (ret!=0) {
		return FI_EPARSE;
	}
	(fltd_new->fltd).net.port = (unsigned short) num; /* valid port found */

	/* initialize socket descriptor */
	ret = sscanf((fltd_new->fltd).net.device, "eth%d", &device);
	assert(ret==1);
	assert(0<=device && device<MAX_ETH_CARDS);
	(fltd_new->fltd).net.fd = FD_ETH(device);

	/* Now find socket descriptor is used for this device. */
	/* If this is an initial fault, fi_.netinfo[] does not yet contain the
	 * information we need. So we mark the appropriate entry for the
	 * routines, who fill in the information, to tell them, to update the
	 * fault description too.
	 * If this is not an initial fault, fi_.netinfo[] may contain the
	 * information we need (if the requested device exists. Whether or not
	 * this is the case is checked by the frontend.)
	 */

#if 0
	for(i=0 ; i<MAX_ETH_CARDS ; i++) {
		if(fi_.netinfo[i].port == (fltd_new->fltd).net.port) {
			if(0 <= fi_.netinfo[i].fd) { /* valid sock. descr. available */
				(fltd_new->fltd).net.fd = fi_.netinfo[i].fd ;
			}else{
				fi_.netinfo[i].fd= -2 ;
			}
			break ;
		}
	}
#endif
	return(0);
}
/***********************************************************************
 * parse_fltd
 ***********************************************************************
 * Input:     int socket  =   socket descriptor for comm. with frontend
 *            fi_ftld_t **fltd_new  =  pointer to new fault descriptions, !=NULL
 *                                     *fltd_new will be allocated
 * Return:    ENOMEM when no memory can be allocated
 *            FI_EPARSE when a parse error occured
 *            FIEOF if end of while was encountered before start of fault descr.
 *            0 if all went ok
 * Effect:    updates fault description list
 * Called By: fi_tick in fi.c
 **********************************************************************/
int
parse_fltd(
	int socket,
	fi_fltd_t **fltd_new)
{
	int ret;
	char *s;
	loff_t num;

	assert(fltd_new != 0);


	*fltd_new = (fi_fltd_t *) calloc(1, sizeof(fi_fltd_t));
	if (*fltd_new == NULL) {
		exit_on_error(ENOMEM);
	}
	assert(*fltd_new != NULL);
	/* we must read a string as first thing, since we have to be able to
	 * parse and recognize the END_OF_FAULTLOAD string.
	 */
	ret = read_string(socket, &s); /* read fault location */
	num = str_to_num(s);
	free(s) ;
	if ( ! ((long)num>0) ) {
		return FI_EPARSE;
	}
	if (num == END_OF_FAULTLOAD) {
		return END_OF_FAULTLOAD ;
	}
	(*fltd_new)->location = (int) num ; /* valid location found */

	ret = read_string(socket, &s); /* read fault type */
	num = str_to_num(s);
	free(s) ;
	if ( ! ((long)num>0) ) {
		return FI_EPARSE;
	}
	assert(FAULT_DURATION != (*fltd_new)->type) ;
	(*fltd_new)->type = (int) num ; /* valid type found */

	ret = read_decnum(socket, &num); /* read fault id */
	if (ret!=0) {
		return FI_EPARSE;
	}
	(*fltd_new)->id = (int) num ; /* valid id */

	switch ((*fltd_new)->location) {
		case FAULT_HD_BLOCKS:
			ret = parse_hd_blocks(socket, *fltd_new);
			break;
		case FAULT_HD_FAIL:
			ret = parse_hd_fail(socket, *fltd_new);
			break;

		case FAULT_RAM:
			ret = parse_ram(socket, *fltd_new);
			break;

		case FAULT_CPU_REGS:
			ret = parse_cpu_regs(socket, *fltd_new);
			break;

		case FAULT_NET_SEND:
		case FAULT_NET_RECV:
			ret = parse_net(socket, *fltd_new);
			break;
		case FAULT_DURATION_END:
			fi_dprintf("fault duration end\n") ;
			/* nothing to do, but is legal fault description */
			break;
		case FAULT_POWER:
			fi_dprintf("fault power\n") ;
			/* nothing to do, but is legal fault description */
			break;

		case FAULT_MMU:
			ret = FI_EPARSE;
			break;

		default:
		fi_eprintf("%s: fi_eparse\n", __FUNCTION__) ;
			ret = FI_EPARSE;
			break;
	}
	return(ret);
}
