/* $Id: tracer.c,v 1.1.1.1 2002/08/16 17:29:27 root Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */
#define DEBUG 0
#define DEBUG_CONTROL_FLOW 0
#define DEBUG_PTRACE 0

#include <assert.h>
#include <errno.h>
#include <sched.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <linux/unistd.h>
#include <sys/ptrace.h>
#include <sys/user.h>
#include <sys/wait.h>

#include "lilo.h"
#include "config.h"
#include "fi_interface.h"

#define EIP	55	/* FIXME */
#define EFL	57	/* FIXME */


enum { USER, KERNEL } mode = KERNEL;
int child_pid;
sigset_t delayed_sigs;

int state[MAX_CPU] = { 0, 0, /*...*/ 0 };
int numcpu;
extern int pids[MAX_CPU];

#if DEBUG_CONTROL_FLOW
#define cfprint(fmt, arg...)	fprintf(stderr, fmt, ##arg)
#else
#define cfprint(fmt, arg...)
#endif

#if DEBUG_CONTROL_FLOW
static void
cfsyscall_call(char *str, int pid, struct user_regs_struct *regs)
{
	fprintf(stderr, "%s: (id=%ld, eip=0x%08lx): ",
			str, (regs->esp - 0x70000000) >> 13, regs->eip);

	switch (regs->orig_eax) {

	case __NR_exit: /* 1 */
		fprintf(stderr, "_exit(%ld)", regs->ebx);
		break;

	case __NR_fork: /* 2 */
		fprintf(stderr, "fork()");
		break;

	case __NR_read: /* 3 */
		fprintf(stderr, "read(%ld, 0x%08lx, %ld)",
				regs->ebx, regs->ecx, regs->edx);
		break;

	case __NR_write: /* 4 */
		fprintf(stderr, "write(%ld, 0x%08lx, %ld)",
				regs->ebx, regs->ecx, regs->edx);
		break;

	case __NR_open: /* 5 */
		fprintf(stderr, "open(0x%08lx, 0x%lx, 0%lx)",
				regs->ebx, regs->ecx, regs->edx);
		break;

	case __NR_close: /* 6 */
		fprintf(stderr, "close(%ld)", regs->ebx);
		break;

	case __NR_execve: /* 11 */ {
		/* FIXME VOSSI */
		fprintf(stderr, "execve(...)");
		break;
	    }

	case __NR_lseek: /* 19 */
		fprintf(stderr, "lseek(%ld, 0x%lx, %ld)",
				regs->ebx, regs->ecx, regs->edx);
		break;

	case __NR_ptrace: /* 26 */ {
		fprintf(stderr, "ptrace(%ld, %ld, 0x%08lx, 0x%08lx)",
				regs->ebx, regs->ecx, regs->edx, regs->esi);
		break;
	    }

	case __NR_dup: /* 41 */ {
		/* FIXME VOSSI */
		fprintf(stderr, "dup(...)");
		break;
	    }

	case __NR_fcntl: /* 55 */ {
		/* FIXME VOSSI */
		fprintf(stderr, "fcntl(...)");
		break;
	    }

	case __NR_gettimeofday: /* 78 */
		fprintf(stderr, "gettimeofday(%p, %p)",
				(char *) regs->ebx, (char *) regs->ecx);
		break;

	case __NR_mmap: /* 90 */ {
		struct args {
			unsigned long addr;
			unsigned long len;
			unsigned long prot;
			unsigned long flags;
			unsigned long fd;
			unsigned long offset;
		} args;

		errno = 0;
		args.addr = ptrace(PTRACE_PEEKDATA, pid, regs->ebx + 0, 0);
		args.len = ptrace(PTRACE_PEEKDATA, pid, regs->ebx + 4, 0);
		args.prot = ptrace(PTRACE_PEEKDATA, pid, regs->ebx + 8, 0);
		args.flags = ptrace(PTRACE_PEEKDATA, pid, regs->ebx + 12, 0);
		args.fd = ptrace(PTRACE_PEEKDATA, pid, regs->ebx + 16, 0);
		args.offset = ptrace(PTRACE_PEEKDATA, pid, regs->ebx + 20, 0);
		if (errno != 0) {
			fprintf(stderr, "mmap(?????)");
		} else {
			fprintf(stderr, "mmap(%p, 0x%lx, 0x%lx, 0x%lx, %ld, 0x%lx)\n",
					(char *) args.addr, args.len, args.prot,
					args.flags, args.fd, args.offset);
		}
		break;
	    }
	case __NR_munmap: /* 91 */
		fprintf(stderr, "munmap(0x%08lx, 0x%08lx)",
				regs->ebx, regs->ecx);
		break;

	case __NR_setitimer: /* 104 */ {
		/* FIXME VOSSI */
		fprintf(stderr, "setitimer(...)");
		break;
	    }

	case __NR_idle: /* 112 */
		fprintf(stderr, "idle()");
		break;

	case __NR_sigreturn: /* 119 */ {
		/* FIXME VOSSI */
		fprintf(stderr, "sigreturn()");
		break;
	    }
	case __NR_clone: /* 120 */ {
		static struct {
			int flag;
			char *name;
		} flag[] = {
			{ CLONE_VM, "CLONE_VM" },
			{ CLONE_FS, "CLONE_FS" },
			{ CLONE_FILES, "CLONE_FILES" },
			{ CLONE_SIGHAND, "CLONE_SIGHAND" },
			{ CLONE_PID, "CLONE_PID" },
			{ CLONE_PTRACE, "CLONE_PTRACE" },
			{ CLONE_VFORK, "CLONE_VFORK" },
			{ 0, (char *) 0 }
		};
		int i;

		fprintf(stderr, "clone(");
		fprintf(stderr, "0x%02x", (int) (regs->ebx & CSIGNAL));
		for (i = 0; flag[i].flag != 0; i++) {
			if (regs->ebx & flag[i].flag) {
				fprintf(stderr, "|");
				fprintf(stderr, "%s", flag[i].name);
			}
		}
		fprintf(stderr, ", ");
		fprintf(stderr, "0x%08lx", regs->ecx);
		fprintf(stderr, ")");
		break;
	    }
	case __NR_sigprocmask: /* 126 */ {
		static char *op[] = {
			"SIG_BLOCK", "SIG_UNBLOCK", "SIG_SETMASK"
		};
		unsigned long mask;	/* FIXME VOSSI */

		errno = 0;
		mask = ptrace(PTRACE_PEEKDATA, pid, regs->ecx + 0, 0);
		if (errno != 0) {
			fprintf(stderr, "sigprocmask(%s, {%s}, %p)",
				(0 <= regs->ebx && regs->ebx < 3)
						? op[regs->ebx] : "?????",
				"?????",
				(char *) regs->edx);
		} else {
			fprintf(stderr, "sigprocmask(%s, {%08lx}, %p)",
				(0 <= regs->ebx && regs->ebx < 3)
						? op[regs->ebx] : "?????",
				mask,
				(char *) regs->edx);
		}
		break;
	    }
	case __NR_rt_sigreturn: /* 173 */ {
		/* FIXME VOSSI */
		fprintf(stderr, "rt_sigreturn()");
		break;
	    }
	case __NR_rt_sigaction: /* 174 */ {
		/* FIXME VOSSI */
		fprintf(stderr, "rt_sigaction(...)");
		break;
	    }
	case __NR_rt_sigsuspend: /* 179 */ {
		unsigned long mask0;	/* FIXME VOSSI */
		unsigned long mask1;	/* FIXME VOSSI */

		fprintf(stderr, "rt_sigsuspend(");
		fprintf(stderr, "{");
		errno = 0;
		mask0 = ptrace(PTRACE_PEEKDATA, pid, regs->ebx + 0, 0);
		mask1 = ptrace(PTRACE_PEEKDATA, pid, regs->ebx + 4, 0);
		if (errno != 0) {
			fprintf(stderr, "????????????????");
		} else {
			fprintf(stderr, "%08lx%08lx", mask1, mask0);
		}
		fprintf(stderr, "}");
		fprintf(stderr, ", ");
		fprintf(stderr, "%ld)", regs->ecx);
		break;
	    }
	case __NR_sigaltstack: /* 186 */
		fprintf(stderr, "sigaltstack(%p, %p)",
				(char *) regs->ebx, (char *) regs->ecx);
		break;

	default:
		fprintf(stderr, "syscall%ld(0x%08lx, 0x%08lx, 0x%08lx, 0x%08lx, 0x%08lx, 0x%08lx)",
				regs->orig_eax,
				regs->ebx, regs->ecx, regs->edx,
				regs->esi, regs->edi, regs->ebp);
	}

	fprintf(stderr, "\n");
}
#else
#define cfsyscall_call(str, pid, regs)
#endif

#if DEBUG_CONTROL_FLOW
static void
cfsyscall_intermediate(char *str, int pid)
{
#if 0
	struct user_regs_struct regs;

	errno = 0;
	ptrace(PTRACE_GETREGS, pid, 0, (unsigned long) &regs);
	assert(errno == 0);

	fprintf(stderr, "%s: (id=%ld, eip=0x%08lx): ",
			str, (regs.esp - 0x70000000) >> 13, regs.eip);

	fprintf(stderr, "working...");

	fprintf(stderr, "\n");
#endif
}
#else
#define cfsyscall_intermediate(str, pid)
#endif

#if DEBUG_CONTROL_FLOW
static void
cfsyscall_return(char *str, int pid)
{
	struct user_regs_struct regs;

	errno = 0;
	ptrace(PTRACE_GETREGS, pid, 0, (unsigned long) &regs);
	assert(errno == 0);

	fprintf(stderr, "%s: (id=%ld, eip=0x%08lx): ",
			str, (regs.esp - 0x70000000) >> 13, regs.eip);

	fprintf(stderr, "return 0x%08lx/%ld", regs.eax, regs.eax);

	fprintf(stderr, "\n");
}
#else
#define cfsyscall_return(str, pid)
#endif

#if DEBUG_CONTROL_FLOW
static void
cfsignal(char *str, int pid, int sig)
{
	struct user_regs_struct regs;

	fprintf(stderr, "%s: signal %d: ", str, sig);

	errno = 0;
	ptrace(PTRACE_GETREGS, pid, 0, (unsigned long) &regs);
	assert(errno == 0);

	fprintf(stderr, "eip=0x%08lx, esp=0x%08lx",
			regs.eip, regs.esp);

	fprintf(stderr, "\n");
}
#else
#define cfsignal(str, pid, sig)
#endif

#if DEBUG_PTRACE
static int
lilo_waitpid(int pid, int *status, int flags)
{
	int ret;

	ret = waitpid(pid, status, flags);

	if (WIFSTOPPED(*status)) {
		fprintf(stderr, "\twaitpid(pid=%d) -> stopped/%d\n",
				pid, WSTOPSIG(*status));
	} else if (WIFSIGNALED(*status)) {
		fprintf(stderr, "\twaitpid(pid=%d) -> signaled/%d\n",
				pid, WTERMSIG(*status));
	} else if (WIFEXITED(*status)) {
		fprintf(stderr, "\twaitpid(pid=%d) -> terminated/%d\n",
				pid, WEXITSTATUS(*status));
	} else {
		assert(0);
	}

	return ret;
}
#else
#define lilo_waitpid(pid, status, flags) waitpid(pid, status, flags)
#endif

#if DEBUG_PTRACE
static long
lilo_ptrace(int request, int pid, unsigned long addr, unsigned long data)
{
	long ret;

	ret = ptrace(request, pid, addr, data);

	switch (request) {
	case PTRACE_PEEKTEXT:
	case PTRACE_PEEKDATA:
		fprintf(stderr, "\tptrace(PEEKDATA, pid=%d, addr=0x%08lx)\n",
				pid, addr);
		break;
	case PTRACE_POKETEXT:
	case PTRACE_POKEDATA:
		fprintf(stderr, "\tptrace(POKEDATA, pid=%d, addr=0x%08lx)\n",
				pid, addr);
		break;
	case PTRACE_GETREGS:
		fprintf(stderr, "\tptrace(GETREGS, pid=%d)\n", pid);
		break;
	case PTRACE_SETREGS:
		fprintf(stderr, "\tptrace(SETREGS, pid=%d)\n", pid);
		break;
	case PTRACE_CONT:
		fprintf(stderr, "\tptrace(CONT, pid=%d, sig=%d)\n",
				pid, (int) data);
		break;
	case PTRACE_SYSCALL:
		fprintf(stderr, "\tptrace(SYSCALL, pid=%d, sig=%d)\n",
				pid, (int) data);
		break;
	case PTRACE_SINGLESTEP:
		fprintf(stderr, "\tptrace(SINGLESTEP, pid=%d, sig=%d)\n",
				pid, (int) data);
		break;
	default:
		fprintf(stderr, "\tptrace(request=%d, pid=%d, addr=0x%08lx, data=0x%08lx) -> 0x%08lx\n",
				request, pid, addr, data, ret);
		break;
	}

	return ret;
}
#else
#define lilo_ptrace(request, pid, addr, data) ptrace(request, pid, addr, data)
#endif

static void
dump_eflags(int pid, unsigned long eflags)
{
	static struct {
		int bit;
		char *name;
	} bitnames[] = {
		{ 21, "ID" },
		{ 20, "VIP" },
		{ 19, "VIF" },
		{ 18, "AC" },
		{ 17, "VM" },
		{ 16, "RF" },
		{ 14, "NT" },
		/* { 13-12, "IOPL" }, */
		{ 11, "OF" },
		{ 10, "DF" },
		{ 9, "IF" },
		{ 8, "TF" },
		{ 7, "SF" },
		{ 6, "ZF" },
		/* { 5, "unused" }, */
		{ 4, "AF" },
		/* { 3, "unused" }, */
		{ 2, "PF" },
		/* { 1, "unused" }, */
		{ 0, "CF" },

		{ -1, (char *) 0 }
	};
	int first;
	int i;

	printf("eflags=0x%08lx (", eflags);
	first = 1;
	for (i = 0; 0 <= bitnames[i].bit; i++) {
		if (eflags & (1 << bitnames[i].bit)) {
			if (first) {
				first = 0;
			} else {
				printf(",");
			}
			printf("%s", bitnames[i].name);
		}
	}
	if (first) {
		first = 0;
	} else {
		printf(",");
	}
	printf("IOPL=%d", (int) (eflags >> 12) & 3);
	printf(")");
}

static void
dump_eip(int pid, unsigned long eip)
{
	unsigned long ret;

	printf("eip=0x%08lx", eip);

	errno = 0;
	ret = ptrace(PTRACE_PEEKTEXT, pid, eip, 0);
	if (errno == 0) {
		printf(" *eip=%08lx ", ret);
	} else {
		printf(" *eip=????????" );
	}
}

static void
dump(char *str, int pid, int sig)
{
	struct user_regs_struct regs;
	unsigned long addr;
	int mcount;
	int ucount;
	int ret;

	/* print string and signal */
	printf("%s:", str);
	if (sig != 0) {
		printf(" got signal %d", sig);
	}
	printf("\n");

	/* print registers */
	ret = ptrace(PTRACE_GETREGS, pid, 0, (unsigned long) &regs);
	assert(0 <= ret);

	printf("\t");
	dump_eip(pid, regs.eip);
	
	printf("esp=%08lx ", regs.esp);
	errno = 0;
	ret = ptrace(PTRACE_PEEKTEXT, pid, regs.esp, 0);
	if (errno == 0) {
		printf("*esp=%08x", ret);
	} else {
		printf("*esp=????????");
	}
	printf("\n");

	printf("\t");
	printf("eax=%08lx ", regs.eax);
	printf("ebx=%08lx ", regs.ebx);
	printf("ecx=%08lx ", regs.ecx);
	printf("edx=%08lx ", regs.edx);
	printf("\n");

	printf("\t");
	printf("esi=%08lx ", regs.esi);
	printf("edi=%08lx ", regs.edi);
	printf("ebp=%08lx ", regs.ebp);
	printf("\n");

	printf("\t");
	printf("orig_eax=%08lx", regs.orig_eax);
	printf(" ");
	dump_eflags(pid, regs.eflags);
	printf("\n");

	/* print number of mounted kernel pages */
	printf("\t");
	mcount = 0;
	ucount = 0;
	for (addr = MEM_BASE + 4096;
	    addr < MEM_BASE + config.mem_size;
	    addr += 4096) {
		errno = 0;
		ret = ptrace(PTRACE_PEEKTEXT, pid, addr, 0);
		if (errno == 0) {
			mcount++;
		} else {
			ucount++;
		}
	}
	if (mcount == 0) {
		printf("kernel NOT accessable");
	} else if (ucount == 0) {
		printf("kernel accessable");
	} else {
		printf("kernel PARTLY accessable (%d/%d)",
				mcount, mcount + ucount);
	}
	printf("\n");
}

static void
die(int pid, int sig)
{
	dump("die", pid, sig);
	kill(pid, SIGKILL);
	printf("exiting\n");
	exit(1);
}

/*
 * Start a new or resume an already started syscall 
 */
static void
sys_cont_wait(int mode, int pid, int contsig, int stopsig, sigset_t *sigs, int status)
{
	int ret;
	int i;

#if DEBUG
	fprintf(stderr, "cont_wait: mode=%d, contsig=%d, stopsig=%d, sigs=%p\n",
			mode, contsig, stopsig, sigs);
#endif
	assert(0 < pid);
	assert(contsig == 0 || (1 <= contsig && contsig < _NSIG));
	assert(1 <= stopsig && stopsig < _NSIG);

	/* Get state index */
	for (i = 0; i < numcpu; i++) {
		if (pids[i] == pid)
			break;
	}
	assert(i < numcpu);

	if (!state[i]) {
		/* Syscall was not started yet... */
		assert(status == 0);

		ret = lilo_ptrace(mode, pid, 0, contsig);
		assert(0 <= ret);

		/* ... now it is */
		state[i] = 1;

		return;
	}

	assert(WIFSTOPPED(status));
	assert(WSTOPSIG(status) == SIGALRM
	    || WSTOPSIG(status) == SIGHUP
	    || WSTOPSIG(status) == SIGIO
	    || WSTOPSIG(status) == SIGTERM
	    || WSTOPSIG(status) == stopsig);

	/* spin for signals up to `stopsig' */
	while (WSTOPSIG(status) != stopsig) {
		assert(WSTOPSIG(status) == SIGALRM
		    || WSTOPSIG(status) == SIGHUP
		    || WSTOPSIG(status) == SIGIO
		    || WSTOPSIG(status) == SIGTERM);

#if DEBUG
		fprintf(stderr, "cont_wait: got signal %d\n", WSTOPSIG(status));
#endif

		if (sigs != (sigset_t *) 0) {
			ret = sigaddset(sigs, WSTOPSIG(status));
			assert(0 <= ret);

			ret = lilo_ptrace(mode, pid, 0, 0);
			assert(0 <= ret);
		} else {
			ret = lilo_ptrace(mode, pid, 0, WSTOPSIG(status));
			assert(0 <= ret);
		}

		ret = lilo_waitpid(pid, &status, WUNTRACED);
		assert(ret == pid);
		assert(WIFSTOPPED(status));
	}
	assert(WSTOPSIG(status) == stopsig);

#if DEBUG
	cfprint("cont_wait: got stopsig (%d)\n", stopsig);
#endif
}

static void
cont_wait(int mode, int pid, int contsig, int stopsig, sigset_t *sigs)
{
	int status;
	int ret;

#if DEBUG
	fprintf(stderr, "cont_wait: mode=%d, contsig=%d, stopsig=%d, sigs=%p\n",
			mode, contsig, stopsig, sigs);
#endif
	assert(0 < pid);
	assert(contsig == 0 || (1 <= contsig && contsig < _NSIG));
	assert(1 <= stopsig && stopsig < _NSIG);

	ret = lilo_ptrace(mode, pid, 0, contsig);
	assert(0 <= ret);

	ret = lilo_waitpid(pid, &status, WUNTRACED);
	assert(ret == pid);
	assert(WIFSTOPPED(status));
	assert(WSTOPSIG(status) == SIGALRM
	    || WSTOPSIG(status) == SIGHUP
	    || WSTOPSIG(status) == SIGIO
	    || WSTOPSIG(status) == SIGTERM
	    || WSTOPSIG(status) == stopsig);

	/* spin for signals up to `stopsig' */
	while (WSTOPSIG(status) != stopsig) {
		assert(WSTOPSIG(status) == SIGALRM
		    || WSTOPSIG(status) == SIGHUP
		    || WSTOPSIG(status) == SIGIO
		    || WSTOPSIG(status) == SIGTERM);

#if DEBUG
		fprintf(stderr, "cont_wait: got signal %d\n", WSTOPSIG(status));
#endif

		if (sigs != (sigset_t *) 0) {
			ret = sigaddset(sigs, WSTOPSIG(status));
			assert(0 <= ret);

			ret = lilo_ptrace(mode, pid, 0, 0);
			assert(0 <= ret);
		} else {
			ret = lilo_ptrace(mode, pid, 0, WSTOPSIG(status));
			assert(0 <= ret);
		}

		ret = lilo_waitpid(pid, &status, WUNTRACED);
		assert(ret == pid);
		assert(WIFSTOPPED(status));
	}
	assert(WSTOPSIG(status) == stopsig);

#if DEBUG
	cfprint("cont_wait: got stopsig (%d)\n", stopsig);
#endif
}

static void
do_map(int pid, struct user_regs_struct *regs)
{
	struct user_regs_struct save_regs;
	int ret;


	save_regs = *regs;

	regs->eip = MEM_BASE + 0x0000;	/* start of mapping code */
	regs->eax = MEM_BASE;
	regs->ebx = config.mem_size;
	regs->ecx = regs->esp & 0xffffe000;	/* FIXME VOSSI */
	ret = lilo_ptrace(PTRACE_SETREGS, pid, 0, (unsigned long) regs);
	assert(0 <= ret);

	/* execute up to breakpoint */
	cont_wait(PTRACE_CONT, pid, 0, SIGTRAP, &delayed_sigs);

	*regs = save_regs;
	ret = lilo_ptrace(PTRACE_SETREGS, pid, 0, (unsigned long) regs);
	assert(0 <= ret);
}

static void
do_unmap(int pid, struct user_regs_struct *regs)
{
	struct user_regs_struct save_regs;
	int ret;


	save_regs = *regs;

	regs->eip = MEM_BASE + 0x0800;	/* start of unmapping code */
	regs->eax = MEM_BASE;
	regs->ebx = config.mem_size;
	regs->ecx = regs->esp & 0xffffe000;	/* FIXME VOSSI */
	ret = lilo_ptrace(PTRACE_SETREGS, pid, 0, (unsigned long) regs);
	assert(0 <= ret);

	/* execute up to breakpoint */
	cont_wait(PTRACE_CONT, pid, 0, SIGTRAP, &delayed_sigs);

	*regs = save_regs;
	ret = lilo_ptrace(PTRACE_SETREGS, pid, 0, (unsigned long) regs);
	assert(0 <= ret);
}


static int
do_exit(void)
{
	cfprint("child exiting\n");

	/* just continue without signal */
	return 0;
}

static int
do_sigreturn(int pid, struct user_regs_struct *regs)
{
	unsigned long eflags;
	unsigned long eip;
	int ret;

	cfsyscall_call("system call (kernel mode)", pid, regs);

	/* 
	 * sigreturn needs special treatment if the TF flag is set:
	 * 	- normal opcode: do a single step
	 * 	- opcode int 3 or int N: continue
	 */
	assert(mode == KERNEL);

	/*
	 * get eflags
	 */
	errno = 0;
	eflags = lilo_ptrace(PTRACE_PEEKDATA, pid,
		regs->esp + EFL * sizeof(int) - sizeof(void*), 0);
	assert(errno == 0);

	/*
	 * check eflags
	 */
	if ((eflags & 0xffc0802a) != 0x00000002) {
		printf("ERROR: bad eflags:\n");
		dump_eflags(pid, eflags);
		printf("\n");
		assert(0);
	}
	/* if ((eflags & 0x003f7600) != 0x00243200) { ... } */	/* IOPL ignored */
								/* AC ignored */
								/* RF ignored */
	if ((eflags & 0x003a4600) != 0x00200200) {
		/* ID set */
		/* VIP cleared */
		/* VIF cleared */
		/* AC set */		/* currently ignored */
		/* VM cleared */
		/* RF cleared */	/* currently ignored */
		/* NT cleared */
		/* IOPL 3 */		/* currently ignored */
		/* DF cleared */
		/* IF set */

		static int count = 0;

		if (count < 10) {
			printf("WARNING: probably bad eflags:\n");
			dump_eflags(pid, eflags);
#if 1
			eip = lilo_ptrace(PTRACE_PEEKDATA, pid,
				regs->esp + EIP * sizeof(int) - sizeof(void*), 0);
			assert(errno == 0);

			printf(" at ");
			dump_eip(pid, eip);
			printf("\n");
#endif
			count++;

		} else if (count == 10) {
			printf("WARNING: more probably bad eflags follow.\n");
			count++;
		}
		eflags &= ~0x100;	/* FIXME */
	}

	/* first of all, finish sigreturn */
	cont_wait(PTRACE_SYSCALL, pid, 0, SIGTRAP, (sigset_t *) 0);

	cfsyscall_intermediate("system call (kernel mode)", pid);

	/* continue up to signal handling */
	cont_wait(PTRACE_SYSCALL, pid, SIGINT, SIGINT, &delayed_sigs);

	cfsyscall_intermediate("system call (kernel mode)", pid);

	if (opt_p) {
		ret = lilo_ptrace(PTRACE_GETREGS, pid, 0, (unsigned long) regs);
		assert(0 <= ret);

		if (regs->eip < MEM_BASE) {
			/*
			 * going from kernel-mode to user-mode
			 */
			do_unmap(pid, regs);
			mode = USER;
		}
	}

	if (eflags & 0x0100) {
		/* try to do single-step */
		cfsyscall_return("system call (kernel mode) single step", pid);
		return -1;
	} else {
		/* continue */
		cfsyscall_return("system call (kernel mode) continue", pid);
		return 0;
	}
}

/*
 * Start a systemcall
 */
static int
start_smsyscall(int pid, struct user_regs_struct *regs)
{
	int i;
	cfsyscall_call("system call (kernel mode)", pid, regs);

	/*
	 * just start system mode syscall
	 */
	if (regs->eax != -ENOSYS) {
		fprintf(stderr, "regs->eax=0x%lx\n", regs->eax);
	}
	assert(regs->eax == -ENOSYS);
	assert(regs->orig_eax != __NR_exit
	    && regs->orig_eax != __NR_rt_sigreturn);

	fi_watch_enter(pid, regs); /* do faultinjection */

	/* Get state index */
	for (i = 0; i < numcpu; i++) {
		if (pids[i] == pid)
			break;
	}
	assert(i < numcpu);

	assert(!state[i]);

	sys_cont_wait(PTRACE_SYSCALL, pid, 0, SIGTRAP, (sigset_t *) 0, 0);

	return -2;
}

/*
 * Finish a systemcall
 */
static int
finish_syscall(int pid, int status)
{
	int i;
	int ret;
	struct user_regs_struct regs;

	/* Get state index */
	for (i = 0; i < numcpu; i++) {
		if (pids[i] == pid)
			break;
	}
	assert(i < numcpu);

	assert(state[i]);

	sys_cont_wait(PTRACE_SYSCALL, pid, 0, SIGTRAP, (sigset_t *) 0, status);

	cfsyscall_intermediate("system call (kernel mode)", pid);

	/* continue up to signal handling */
	cont_wait(PTRACE_SYSCALL, pid, SIGINT, SIGINT, &delayed_sigs);

	/* syscall is finished */
	state[i] = 0;

	cfsyscall_intermediate("system call (kernel mode)", pid);
	
	ret = lilo_ptrace(PTRACE_GETREGS, pid, 0, (unsigned long) &regs);
	assert (0 <= ret);

	fi_watch_return(pid, &regs); /* do faultinjection */

	cfsyscall_return("system call (kernel mode)", pid);

	return 0;
}


static int
do_umint3(int pid)
{
	cfprint("child doing int3 instruction\n");

	/* continue with SIGUSR2 */
	return SIGUSR2;
}

static int
do_umsyscall(int pid, struct user_regs_struct *regs) 
{
	struct  user_regs_struct saved_regs;
	int ret;


	cfsyscall_call("system call (user mode)", pid, regs);

	/* 
	 * continue with getpid() (to finish the real systemcall),
	 * then continue with SIGUSR1 to handle user mode syscall
	 */
	assert(regs->eax == -ENOSYS);

	/* save registers */
	saved_regs = *regs;

	/* continue doing getpid() */
	regs->orig_eax = __NR_getpid;
	regs->eax = __NR_getpid;
	ret = lilo_ptrace(PTRACE_SETREGS, pid, 0, (unsigned long) regs);
	assert(0 <= ret);

	/* do getpid() */
	cont_wait(PTRACE_SYSCALL, pid, 0, SIGTRAP, (sigset_t *) 0);

	cfsyscall_intermediate("system call (user mode)", pid);

	/* restore registers */
	saved_regs.eax = saved_regs.orig_eax;
	ret = lilo_ptrace(PTRACE_SETREGS, pid, 0, (unsigned long) &saved_regs);
	assert(0 <= ret);

	/* continue up to signal handling */
	cont_wait(PTRACE_SYSCALL, pid, SIGINT, SIGINT, &delayed_sigs);

	cfsyscall_intermediate("system call (user mode)", pid);

	cfsyscall_return("system call (user mode)", pid);

	/* continue with SIGUSR1 */
	return SIGUSR1;
}


static int
do_fault_injection(int pid)
{
	cfprint("doing fault injection\n");

	fi_do();

	/* continue without signal */
	return 0;
}

static int
do_signal(int pid, int sig)
{
	cfsignal("child receives signal", pid, sig);

	/* continue with signal */
	return sig;
}

static int
do_syscall(int pid)
{
	struct user_regs_struct regs;
	int cont;
	int ret;

	ret = lilo_ptrace(PTRACE_GETREGS, pid, 0, (unsigned long) &regs);
	assert(0 <= ret);

	/* security hole!!! FIXME VOSSI */
	/*
	 * Should make sure that we're running in kernel mode.
	 * Otherwise `regs.orig_eax == 0' should be disallowed!
	 */
	if (regs.orig_eax == 0
	 || (regs.orig_eax == __NR_rt_sigreturn && MEM_BASE <= regs.eip)) {
		/*
		 * system mode system call
		 */
		assert(regs.eax == -ENOSYS);	/* XXX */

		/* reset registers */
		if (regs.orig_eax == 0) {
			regs.orig_eax = regs.edi >> 16;
			regs.edi &= 0x0000ffff;
			ret = lilo_ptrace(PTRACE_SETREGS, pid,
					0, (unsigned long) &regs);
			assert(0 <= ret);
		}

		if (regs.orig_eax == __NR_exit) {
			/* _exit called */
			cont = do_exit();
			assert(cont == 0);

		} else if (regs.orig_eax == __NR_rt_sigreturn) {
			/* rt_sigreturn called */
			cont = do_sigreturn(pid, &regs);
			assert(cont == 0 || cont == -1);

		} else  {
			/* start system call */
			cont = start_smsyscall(pid, &regs);
			assert(cont == -2);
		}

	} else {
		/*
		 * user mode system call
		 */
		unsigned long opcode;

		errno = 0;
		opcode = lilo_ptrace(PTRACE_PEEKDATA, pid, regs.eip - 2, 0);
		assert(errno == 0);

		assert((opcode & 0x0000ffff) == 0x000080cd
		    || (opcode & 0x0000ff00) == 0x0000cc00);

		/* was trap caused by int 80 or int 3? */
		if ((opcode & 0x0000ff00) == 0x0000cc00) {
			cont = do_umint3(pid);
			assert(cont == SIGUSR2);

		} else {
			cont = do_umsyscall(pid, &regs);
			assert(cont == SIGUSR1);
		}
	}

	return cont;
}


/* when tracer receives a SIGALRM because timer for faultinjection runs out, it
 * does not inject fault right away but instead sends virtual machine a SIGQUIT.
 * Then, when virtual machine stops because of a SIGQUIT, the fault is finally
 * injected. */
/* Since now we want to inject faults when frontend sends new fault via socket,
 * do the same */
static void
sigio_handler(int sig)
{
	assert(sig == SIGIO);

	if (fi_io()) {
		kill(child_pid, SIGQUIT);
	}
}

static void
sighup_handler(int sig)
{
	assert(sig == SIGHUP);

	kill(child_pid, SIGHUP);
}

static void
sigterm_handler(int sig)
{
	assert(sig == SIGTERM);

	kill(child_pid, SIGTERM);
}

int
tracer(int ncpu)
{
	sigset_t mask;
	sigset_t maskold;
	struct sigaction saio;
	struct sigaction saioold;
	struct sigaction sahup;
	struct sigaction sahupold;
	struct sigaction saterm;
	struct sigaction satermold;
	int status;
	int acpu;
	int cont;
	int pid;
	int ret;
	int i;


	child_pid = pids[0];
	acpu = ncpu;
	numcpu = ncpu;

	sigemptyset(&delayed_sigs);

	/*
	 * initialize signal handling
	 */
	ret = sigemptyset(&mask);
	assert(0 <= ret) ;
	ret = sigaddset(&mask, SIGIO); /* now for faultinjection */
	assert(0 <= ret) ;
	ret = sigaddset(&mask, SIGHUP);
	assert(0 <= ret) ;
	ret = sigaddset(&mask, SIGTERM);
	assert(0 <= ret) ;
	ret = sigprocmask(SIG_SETMASK, &mask, &maskold) ;
	assert(0 <= ret) ;

	saio.sa_handler = sigio_handler;
	saio.sa_flags = 0;
	ret = sigemptyset(&saio.sa_mask);
	assert(0 <= ret);
	ret = sigaction(SIGIO, &saio, &saioold);
	assert(0 <= ret);

	sahup.sa_handler = sighup_handler;
	sahup.sa_flags = 0;
	ret = sigemptyset(&sahup.sa_mask);
	assert(0 <= ret);
	ret = sigaction(SIGHUP, &sahup, &sahupold);
	assert(0 <= ret);

	saterm.sa_handler = sigterm_handler;
	saterm.sa_flags = 0;
	ret = sigemptyset(&saterm.sa_mask);
	assert(0 <= ret);
	ret = sigaction(SIGTERM, &saterm, &satermold);
	assert(0 <= ret);

	/*
	 * wait for children
	 */
	for (i = 0; i < ncpu; i++) {
		ret = lilo_waitpid(pids[i], &status, WUNTRACED);
		assert(ret == pids[i]);

		ret = lilo_ptrace(PTRACE_SYSCALL, pids[i], 0, 0);
		assert(0 <= ret);
	}
	
	for (;;) {
		/* wait for system call or signal */
		ret = sigprocmask(SIG_SETMASK, &maskold, &mask) ;
		assert(0 <= ret) ;
		while ((pid = lilo_waitpid(0, &status, WUNTRACED)) < 0) {
			assert(errno == EINTR);
		}
		assert(pid != 0);

		child_pid = pid;

		/* Get state index */
		for (i = 0; i < ncpu; i++) {
			if (pids[i] == pid) break;
		}
		assert(i < ncpu);

		ret = sigprocmask(SIG_SETMASK, &mask, &maskold) ;
		assert(0 <= ret);

		if (WIFSIGNALED(status) || WIFEXITED(status)) {
			/* Only leave tracer if all CPUs have terminated */
			if (--acpu == 0)
				break;
			else
				continue;
		}

		assert(WIFSTOPPED(status));

		switch (WSTOPSIG(status)) {
		case SIGQUIT:
			/*
			 * got SIGQUIT signal
			 * want to inject a fault
			 */
			cont = do_fault_injection(pid);
			break;

		case SIGTRAP:
			/*
			 * system call or breakpoint
			 */
			if (state[i])
				cont = finish_syscall(pid, status);
			else
				cont = do_syscall(pid);
			break;

		default:
			/*
			 * got non-SIGQUIT/SIGTRAP signal
			 */
			cont = do_signal(pid, WSTOPSIG(status));
			break;
		}

		if (cont == -1) {
			/* do single-step */
			struct user_regs_struct testregs;
			unsigned long opcode;
			int status;

			ret = lilo_ptrace(PTRACE_GETREGS, pid, 0,
					(unsigned long) &testregs);
			assert(0 <= ret);

			errno = 0;
			opcode = lilo_ptrace(PTRACE_PEEKDATA, pid,
					testregs.eip & ~3, (long) &opcode);
			if (errno != 0) opcode = 0x00000000;
			opcode >>= 8 * (testregs.eip & 3);
			opcode &= 0x000000ff;

			if (opcode == 0x000000cc || opcode == 0x000000cd) {
				/* int 3 or int N */
				cont = 0;

			} else {
				/* normal opcode */
				ret = lilo_ptrace(PTRACE_SINGLESTEP, pid, 0, 0);
				assert(0 <= ret);

				ret = lilo_waitpid(pid, &status, WUNTRACED);
				assert(ret == pid);
				assert(WIFSTOPPED(status));

				if (WSTOPSIG(status) != SIGTRAP) {
					cont = WSTOPSIG(status);
				} else {
					cont = SIGPIPE;
				}
			}
		}

		if (cont == 0) {
			/* continue without signal */
			ret = lilo_ptrace(PTRACE_SYSCALL, pid, 0, 0);
			assert(0 <= ret);

		} else if (cont > 0) {
			if (mode == USER && opt_p) {
				/*
				 * going from user-mode to kernel-mode
				 */
				struct user_regs_struct regs;

				/* write signal info on kernel stack */
				/* wait for process calling signal handler */
				cont_wait(PTRACE_SYSCALL, pid, cont, SIGSEGV, (sigset_t *) 0);

				/* map kernel pages */
				ret = lilo_ptrace(PTRACE_GETREGS, pid, 0,
						(unsigned long) &regs);
				assert(0 <= ret);

				do_map(pid, &regs);
				mode = KERNEL;

				/* continue processing in kernel space */
				ret = lilo_ptrace(PTRACE_SYSCALL, pid, 0, 0);
				assert(0 <= ret);

			} else {
				/*
				 * continuing in kernel-mode
				 */
				/* continue with signal */
				ret = lilo_ptrace(PTRACE_SYSCALL, pid, 0, cont);
				assert(0 <= ret);
			}
		}

		/*
		 * deliver delayed signals
		 */
		if (sigismember(&delayed_sigs, SIGALRM)) {
			ret = kill(pid, SIGALRM);
			assert(0 <= ret);
			sigdelset(&delayed_sigs, SIGALRM);

		} else if (sigismember(&delayed_sigs, SIGHUP)) {
			ret = kill(pid, SIGHUP);
			assert(0 <= ret);
			sigdelset(&delayed_sigs, SIGHUP);

		} else if (sigismember(&delayed_sigs, SIGIO)) {
			ret = kill(pid, SIGIO);
			assert(0 <= ret);
			sigdelset(&delayed_sigs, SIGIO);

		} else if (sigismember(&delayed_sigs, SIGTERM)) {
			ret = kill(pid, SIGTERM);
			assert(0 <= ret);
			sigdelset(&delayed_sigs, SIGTERM);
		}
	}

	if (WIFEXITED(status)) {
		printf("Program terminated with exit status %d\n",
				WEXITSTATUS(status));
	} else if (WIFSIGNALED(status)) {
		printf("Programm died on signal %d\n", WTERMSIG(status));
	}

	/*
	 * reset signal handling
	 */
	ret = sigaction(SIGTERM, &satermold, (struct sigaction *) 0);
	assert(0 <= ret);

	ret = sigaction(SIGHUP, &sahupold, (struct sigaction *) 0);
	assert(0 <= ret);

	ret = sigaction(SIGIO, &saioold, (struct sigaction *) 0);
	assert(0 <= ret);

	ret = sigprocmask(SIG_SETMASK, &maskold, (sigset_t *) 0) ;
	assert(0 <= ret) ;

	if (WIFEXITED(status)) {
		return WEXITSTATUS(status);
	} else {
		return 1;
	}
}
